from pkilint import validation
from pkilint.adobe import asn1


class AdobeTimestampValidator(validation.Validator):
    VALIDATION_INVALID_GENERALNAME_TYPE = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'adbe.invalid_timestamp_location_type'
    )

    def __init__(self):
        super().__init__(validations=[self.VALIDATION_INVALID_GENERALNAME_TYPE], pdu_class=asn1.AdobeTimestamp)

    def validate(self, node):
        gn = node.children['location']

        gn_type, _ = gn.child

        if gn_type != 'uniformResourceIdentifier':
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_INVALID_GENERALNAME_TYPE,
                f'Invalid Adobe timestamp location type: {gn_type}'
            )



from pkilint import validation
from pkilint.pkix import time


class CrlSaneValidityPeriodValidator(time.SaneValidityPeriodValidator):
    VALIDATION_NEGATIVE_VALIDITY_PERIOD = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'pkix.crl_negative_validity_period'
    )

    def __init__(self):
        super().__init__(
            end_validity_node_retriever=lambda n: n.navigate('^.nextUpdate'),
            path='certificateList.tbsCertificateList.thisUpdate',
            validation=self.VALIDATION_NEGATIVE_VALIDITY_PERIOD
        )



import base64
import functools
import re
import sys

from pkilint.pkix.certificate import RFC5280Certificate
from pkilint.pkix.crl import RFC5280CertificateList
from pkilint.pkix.ocsp import RFC6960OCSPResponse


def _create_ascii_armor(document_kind: str):
    document_kind = document_kind.upper()

    return f'-----BEGIN {document_kind}-----', f'-----END {document_kind}-----'


def _create_pem_re(ascii_armor=('', '')) -> re.Pattern:
    return re.compile(f'^\\s*{ascii_armor[0]}(?P<pem>.+){ascii_armor[1]}\\s*$', re.DOTALL)


_CERTIFICATE_PEM_REGEX = _create_pem_re(_create_ascii_armor('CERTIFICATE'))
_CRL_PEM_REGEX = _create_pem_re(_create_ascii_armor('X509 CRL'))
_GENERIC_BASE64_REGEX = _create_pem_re()

_DOCUMENT_CLS_TO_PEM_REGEX = {
    RFC5280Certificate: _CERTIFICATE_PEM_REGEX,
    RFC5280CertificateList: _CRL_PEM_REGEX,
}


def _convert_pem_str_to_der(regex: re.Pattern, pem_text: str) -> bytes:
    m = regex.match(pem_text)

    if m is None:
        raise ValueError('Invalid PEM text')

    b64_text = m.group('pem')

    return base64.b64decode(b64_text)


def _convert_pem_bytes_to_der(regex: re.Pattern, pem_text: bytes) -> bytes:
    return _convert_pem_str_to_der(regex, pem_text.decode())


def _load_der_document(document_cls, substrate: bytes, document_name: str = None,
                       substrate_source: str = None, parent=None):
    if not substrate.startswith(b'\x30'):
        raise ValueError('Substrate is not DER-encoded')

    doc = document_cls(substrate_source, substrate, document_name, parent)
    doc.decode()

    return doc


def _load_pem_document(document_cls, substrate: str, document_name: str = None,
                       substrate_source: str = None, parent=None):
    regex = _DOCUMENT_CLS_TO_PEM_REGEX.get(document_cls, _GENERIC_BASE64_REGEX)

    der = _convert_pem_str_to_der(regex, substrate)

    return _load_der_document(document_cls, der, document_name, substrate_source, parent)


def _load_pem_file(document_cls, f, document_name: str = None, substrate_source: str = None, parent=None):
    data = f.read()

    regex = _DOCUMENT_CLS_TO_PEM_REGEX.get(document_cls, _GENERIC_BASE64_REGEX)

    if isinstance(data, bytes):
        der = _convert_pem_bytes_to_der(regex, data)
    else:
        der = _convert_pem_str_to_der(regex, data)

    return _load_der_document(document_cls, der, document_name, substrate_source, parent)


def _load_der_file(document_cls, f, document_name: str = None, substrate_source: str = None, parent=None):
    data = f.read()

    return _load_der_document(document_cls, data, document_name, substrate_source, parent)


_this_module = sys.modules[__name__]

for doc_name, doc_cls in [
    ('certificate', RFC5280Certificate),
    ('crl', RFC5280CertificateList),
    ('ocsp_response', RFC6960OCSPResponse),
]:
    for func in [
        _load_der_file,
        _load_pem_file,
        _load_der_document,
        _load_pem_document,
    ]:
        if func == _load_der_file:
            func_name = f'load_der_{doc_name}_file'
        elif func == _load_pem_file:
            func_name = f'load_pem_{doc_name}_file'
        elif func == _load_der_document:
            func_name = f'load_der_{doc_name}'
        elif func == _load_pem_document:
            func_name = f'load_pem_{doc_name}'
        else:
            raise ValueError(f'Unknown function: {func}')

        setattr(
            _this_module,
            func_name,
            functools.partial(func, doc_cls)
        )


def _load_document(der_loader, pem_loader, substrate, name=None, substrate_source=None, parent=None):
    if isinstance(substrate, str):
        return pem_loader(substrate, name, substrate_source, parent)
    elif isinstance(substrate, bytes):
        try:
            return der_loader(substrate, name, substrate_source, parent)
        except ValueError:
            pem_str = substrate.decode()

            return pem_loader(pem_str, name, substrate_source, parent)
    else:
        data = substrate.read()

        return _load_document(der_loader, pem_loader, data, name, substrate_source, parent)


def load_certificate(io, substrate_source, name=None, parent=None):
    return _load_document(getattr(_this_module, 'load_der_certificate'), getattr(_this_module, 'load_pem_certificate'),
                          io, name, substrate_source, parent)


def load_crl(io, substrate_source, name=None, parent=None):
    return _load_document(getattr(_this_module, 'load_der_crl'), getattr(_this_module, 'load_pem_crl'),
                          io, name, substrate_source, parent)


def load_ocsp_response(io, substrate_source, name=None, parent=None):
    return _load_document(getattr(_this_module, 'load_der_ocsp_response'),
                          getattr(_this_module, 'load_pem_ocsp_response'), io, name, substrate_source, parent)



import enum
import typing


@enum.unique
class RegistrationSchemeCountryIdentifierType(enum.IntEnum):
    NONE = 0
    XG = 1
    ISO3166 = 2


class RegistrationSchemeNamingConvention(typing.NamedTuple):
    country_identifier_type: RegistrationSchemeCountryIdentifierType
    allow_state_province: bool
    require_registration_reference: bool


REGISTRATION_SCHEMES = {
    'NTR': RegistrationSchemeNamingConvention(RegistrationSchemeCountryIdentifierType.ISO3166, True, True),
    'VAT': RegistrationSchemeNamingConvention(RegistrationSchemeCountryIdentifierType.ISO3166, False, True),
    'PSD': RegistrationSchemeNamingConvention(RegistrationSchemeCountryIdentifierType.ISO3166, False, True),
}



import operator
from typing import Mapping, Tuple

from cryptography import x509
from dateutil.relativedelta import relativedelta
from pyasn1.type import univ
from pyasn1.type.univ import ObjectIdentifier
from pyasn1_alt_modules import rfc8398, rfc5280, rfc4262

import pkilint.adobe.asn1 as adobe_asn1
import pkilint.cabf.cabf_extension
import pkilint.cabf.smime.smime_extension
import pkilint.common
import pkilint.pkix.certificate
from pkilint import validation, cabf, document, etsi
from pkilint.adobe import adobe_validator
from pkilint.cabf import cabf_extension, cabf_key, cabf_name
from pkilint.cabf.smime import (
    smime_constants, smime_name, smime_key, smime_extension
)
from pkilint.cabf.smime.smime_constants import Generation
from pkilint.iso import lei
from pkilint.msft import asn1 as microsoft_asn1
from pkilint.msft import msft_name
from pkilint.pkix import certificate, time
from pkilint.pkix.general_name import OTHER_NAME_MAPPINGS as PKIX_OTHERNAME_MAPPINGS

OTHER_NAME_MAPPINGS = {
    **PKIX_OTHERNAME_MAPPINGS,
    rfc8398.id_on_SmtpUTF8Mailbox: rfc8398.SmtpUTF8Mailbox(),
    microsoft_asn1.id_on_UserPrincipalName: microsoft_asn1.UserPrincipalName(),
}


def determine_validation_level_and_generation(cert,
                                              config: Mapping[univ.ObjectIdentifier,
                                              Tuple[
                                                  smime_constants.ValidationLevel, smime_extension.Generation]] = None):
    crypto_cert = cert.cryptography_object

    try:
        ext = crypto_cert.extensions.get_extension_for_oid(x509.OID_CERTIFICATE_POLICIES)

        oids = {ObjectIdentifier(pi.policy_identifier.dotted_string) for pi in ext.value}

        for v in smime_constants.ValidationLevel:
            for g in smime_constants.Generation:
                oid = smime_constants.get_policy_oid(v, g)

                if oid in oids:
                    return v, g

        if config is not None:
            for o in oids:
                v_g = config.get(o)

                if v_g is not None:
                    return v_g

    except x509.ExtensionNotFound:
        pass

    return None


def _has_subject_attr(cert, attr):
    return any(cert.get_subject_attributes_by_type(attr))


def _get_first_subject_attr_dirstring_value(cert, attr, attr_asn1_cls):
    attrs = cert.get_subject_attributes_by_type(attr)

    if any(attrs):
        attr, _ = attrs[0]

        attr_value_pdu = attr.children['value'].pdu

        decoded_value = document.decode_substrate(cert, attr_value_pdu, attr_asn1_cls())

        # assume DirectoryString
        _, attr_value_choice_value = decoded_value.child

        return str(attr_value_choice_value.pdu)
    else:
        return None


def guess_validation_level_and_generation(cert,
                                          config: Mapping[univ.ObjectIdentifier,
                                          Tuple[smime_constants.ValidationLevel, smime_extension.Generation]] = None):
    v_g = determine_validation_level_and_generation(cert, config)

    if v_g is not None:
        return v_g

    # assume Legacy generation
    g = smime_constants.Generation.LEGACY

    o = _get_first_subject_attr_dirstring_value(cert, rfc5280.id_at_organizationName, rfc5280.X520OrganizationName)
    has_o = o is not None
    cn = _get_first_subject_attr_dirstring_value(cert, rfc5280.id_at_commonName, rfc5280.X520CommonName)
    has_cn = cn is not None
    has_natural_name = _has_subject_attr(cert, rfc5280.id_at_surname) or _has_subject_attr(
        cert, rfc5280.id_at_givenName)

    if has_o and (has_natural_name or (has_cn and o != cn and '@' not in cn)):
        v = smime_constants.ValidationLevel.SPONSORED
    elif has_o:
        v = smime_constants.ValidationLevel.ORGANIZATION
    elif has_natural_name:
        v = smime_constants.ValidationLevel.INDIVIDUAL
    else:
        v = smime_constants.ValidationLevel.MAILBOX

    return v, g


_SMIME_EXTENSION_MAPPINGS = {
    **cabf.EXTENSION_MAPPINGS,
    **lei.EXTENSION_MAPPINGS,
    **rfc4262.certificateExtensionsMap,
    **adobe_asn1.EXTENSION_MAPPINGS,
}


def create_decoding_validators():
    return pkilint.pkix.certificate.create_decoding_validators(
        cabf.NAME_ATTRIBUTE_MAPPINGS,
        _SMIME_EXTENSION_MAPPINGS,
        [certificate.create_other_name_decoder(OTHER_NAME_MAPPINGS),
         certificate.create_qc_statements_decoder(etsi.ETSI_QC_STATEMENTS_MAPPINGS)]
    )


def create_spki_validation_container():
    return validation.ValidatorContainer(
        validators=[
            smime_key.SmimeAllowedPublicKeyAlgorithmEncodingValidator(
                path='certificate.tbsCertificate.subjectPublicKeyInfo.algorithm'
            ),
            cabf_key.RsaKeyValidator(),
            cabf_key.EcdsaKeyValidator(),
            smime_key.GmailAllowedModulusLengthValidator(),
        ],
        path='certificate.tbsCertificate.subjectPublicKeyInfo'
    )


def create_extensions_validator_container(validation_level, generation):
    return certificate.create_extensions_validator_container(
        [
            smime_extension.RequiredPolicyIdentifierValidator(validation_level, generation),
            smime_extension.CertificatePoliciesPresenceValidator(),
            smime_extension.ExtendedKeyUsagePresenceValidator(),
            smime_extension.CabfSmimeKeyUsagePresenceValidator(),
            cabf_extension.AuthorityInformationAccessPresenceValidator(
                validation.ValidationFindingSeverity.WARNING
            ),
            smime_extension.CrlDistributionPointPresenceValidator(),
            smime_extension.SubjectAlternativeNamePresenceValidator(),
            smime_extension.AllowedExtendedKeyUsageValidator(generation),
            smime_extension.AllowedKeyUsageValidator(generation),
            smime_extension.EndEntityValidator(),
            cabf_extension.CpsUriHttpValidator(),
            cabf_extension.AuthorityInformationAccessContainsHttpUriValidator(),
            cabf_extension.CrlDpContainsHttpUriValidator(),
            smime_extension.SubjectAlternativeNameContainsEmailAddressValidator(),
            smime_extension.SubjectAlternativeNameProhibitedGeneralNameTypesValidator(generation),
            smime_extension.CabfSmimeKeyUsageCriticalityValidator(),
            smime_extension.GmailAuthorityInfoAccessCaIssuersValidator(),
            msft_name.UserPrincipalNameSyntaxValidator(),
            smime_extension.AllowedCrldpFullNameValidator(generation),
            smime_extension.SmimeUserNoticeValidator(),
            smime_extension.AllowedAiaUriSchemeValidator(generation),
            smime_extension.LeiCriticalityValidator(),
            smime_extension.LeiRoleCriticalityValidator(),
            smime_extension.LeiPresenceValidator(validation_level),
            smime_extension.LeiRolePresenceValidator(validation_level),
            lei.LeiExtensionValueSyntaxValidator(),
            cabf_extension.CertificatePoliciesCriticalityValidator(),
            cabf_extension.CabfCrlDpCriticalityValidator(),
            cabf_extension.CabfAuthorityKeyIdentifierValidator(),
            smime_extension.SubjectDirectoryAttributesPresenceValidator(validation_level, generation),
            smime_extension.QCStatementsCriticalityValidator(),
            cabf_name.GeneralNameUriInternalDomainNameValidator(),
            cabf_name.GeneralNameRfc822NameInternalDomainNameValidator(),
            cabf_name.SmtpUtf8MailboxInternalDomainNameValidator(),
            cabf_name.UriInternalDomainNameValidator(pdu_class=rfc5280.CPSuri),
            adobe_validator.AdobeTimestampValidator(),
            smime_extension.AdobeTimestampCriticalityValidator(),
            smime_extension.AdobeTimestampPresenceValidator(generation),
            smime_extension.AdobeArchiveRevInfoCriticalityValidator(),
            smime_extension.AdobeArchiveRevInfoPresenceValidator(generation),
        ]
    )


def create_validity_validators(generation):
    days = 1185 if generation == Generation.LEGACY else 825

    threshold_error = (
        operator.le,
        relativedelta(days=days),
        validation.ValidationFinding(
            validation.ValidationFindingSeverity.ERROR,
            f'cabf.smime.certificate_validity_period_exceeds_{days}_days'
        )
    )

    threshold_warning = (
        operator.le,
        relativedelta(days=days - 1, hours=23, minutes=59, seconds=59),
        validation.ValidationFinding(
            validation.ValidationFindingSeverity.WARNING,
            'cabf.smime.certificate_validity_period_at_maximum'
        )
    )
    return [
        time.ValidityPeriodThresholdsValidator(
            path='certificate.tbsCertificate.validity.notBefore',
            end_validity_node_retriever=lambda n: n.navigate('^.notAfter'),
            inclusive_second=True,
            validity_period_thresholds=[threshold_error, threshold_warning]
        )
    ]


def create_subscriber_validators(validation_level, generation):
    return [
        smime_name.create_subscriber_certificate_subject_validator_container(validation_level, generation),
        create_spki_validation_container(),
        certificate.create_issuer_validator_container(
            []
        ),
        certificate.create_validity_validator_container(
            create_validity_validators(generation)
        ),
        create_extensions_validator_container(validation_level, generation),
        smime_key.SmimeAllowedSignatureAlgorithmEncodingValidator(
            path='certificate.tbsCertificate.signature'
        ),
        smime_key.SmimeAllowedSignatureAlgorithmEncodingValidator(
            path='certificate.signatureValue'
        ),
        cabf_extension.CabfExtensionsPresenceValidator(),
    ]



from pyasn1_alt_modules import rfc5280

from pkilint import validation
from pkilint.itu import bitstring


# BR 7.1.2.10.7
class CaKeyUsageValidator(validation.Validator):
    VALIDATION_CA_CERT_REQUIRED_BIT_MISSING = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'cabf.ca_certificate_required_ku_missing'
    )

    VALIDATION_CA_CERT_PROHIBITED_BIT_PRESENT = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'cabf.ca_certificate_prohibited_ku_present'
    )

    VALIDATION_CA_CERT_NO_DIG_SIG = validation.ValidationFinding(
        validation.ValidationFindingSeverity.NOTICE,
        'cabf.ca_certificate_no_digital_signature_bit'
    )

    _PROHIBITED_KUS = {
        str(n) for n in rfc5280.KeyUsage.namedValues
    } - {'digitalSignature', 'keyCertSign', 'cRLSign'}

    def __init__(self):
        super().__init__(
            validations=[
                self.VALIDATION_CA_CERT_REQUIRED_BIT_MISSING,
                self.VALIDATION_CA_CERT_PROHIBITED_BIT_PRESENT,
                self.VALIDATION_CA_CERT_NO_DIG_SIG,
            ],
            pdu_class=rfc5280.KeyUsage
        )

    def validate(self, node):
        if not bitstring.has_named_bit(node, 'keyCertSign'):
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_CA_CERT_REQUIRED_BIT_MISSING,
                'keyCertSign not asserted'
            )
        if not bitstring.has_named_bit(node, 'cRLSign'):
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_CA_CERT_REQUIRED_BIT_MISSING,
                'cRLSign not asserted'
            )
        prohibited_kus_asserted = [k for k in self._PROHIBITED_KUS if bitstring.has_named_bit(node, k)]

        if any(prohibited_kus_asserted):
            prohibited_kus_str = ', '.join(prohibited_kus_asserted)
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_CA_CERT_PROHIBITED_BIT_PRESENT,
                f'Prohibited KUs present: {prohibited_kus_str}'
            )

        if not bitstring.has_named_bit(node, 'digitalSignature'):
            raise validation.ValidationFindingEncountered(self.VALIDATION_CA_CERT_NO_DIG_SIG)


class CaBasicConstraintsValidator(validation.Validator):
    VALIDATION_CA_BIT_NOT_SET = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'cabf.serverauth.ca_basic_constraints_ca_bit_not_set'
    )

    def __init__(self):
        super().__init__(validations=self.VALIDATION_CA_BIT_NOT_SET, pdu_class=rfc5280.BasicConstraints)

    def validate(self, node):
        if not bool(node.children['cA'].pdu):
            raise validation.ValidationFindingEncountered(self.VALIDATION_CA_BIT_NOT_SET)



import ipaddress
from typing import List

from cryptography import x509
from pyasn1_alt_modules import rfc5280, rfc4985, rfc6962

import pkilint.cabf.cabf_name
import pkilint.cabf.serverauth.serverauth_extension
import pkilint.cabf.serverauth.serverauth_name
import pkilint.cabf.serverauth.serverauth_subscriber
import pkilint.common
from pkilint import validation, cabf, etsi
from pkilint.cabf import cabf_key, cabf_name, cabf_extension, cabf_ca
from pkilint.cabf.serverauth import (
    serverauth_name, serverauth_extension, serverauth_constants,
    serverauth_key, serverauth_root, serverauth_ca, serverauth_ocsp, serverauth_cross_ca, serverauth_finding_filter
)
from pkilint.pkix import name, certificate

OTHER_NAME_MAPPINGS = rfc4985.otherNamesMap.copy()


def _has_full_name_constraints(crypto_cert: x509.Certificate):
    has_dirname_constraint = False
    has_ipv4_constraint = False
    has_ipv6_constraint = False
    has_dnsname_constraint = False

    try:
        ext = crypto_cert.extensions.get_extension_for_oid(x509.OID_NAME_CONSTRAINTS)

        gns = (
            [] if ext.value.permitted_subtrees is None else ext.value.permitted_subtrees +
            [] if ext.value.excluded_subtrees is None else ext.value.excluded_subtrees
        )

        for gn in gns:
            if isinstance(gn, x509.DNSName):
                has_dnsname_constraint = True
            elif isinstance(gn, x509.DirectoryName):
                has_dirname_constraint = True
            elif isinstance(gn.value, ipaddress.IPv4Network):
                has_ipv4_constraint = True
            elif isinstance(gn.value, ipaddress.IPv6Network):
                has_ipv6_constraint = True

        return all((has_dirname_constraint, has_ipv4_constraint, has_ipv6_constraint, has_dnsname_constraint))

    except x509.ExtensionNotFound:
        return False


def _get_intermediate_ca_type(crypto_cert: x509.Certificate):
    try:
        ext = crypto_cert.extensions.get_extension_for_oid(x509.OID_EXTENDED_KEY_USAGE)

        ekus = set(ext.value)
    except x509.ExtensionNotFound:
        # assume serverauth
        ekus = {x509.OID_SERVER_AUTH}

    if x509.ObjectIdentifier(str(rfc6962.id_kp_precertificateSigning)) in ekus:
        return serverauth_constants.CertificateType.PRECERT_SIGNING_CA
    elif x509.OID_SERVER_AUTH in ekus or x509.ObjectIdentifier(str(rfc5280.anyExtendedKeyUsage)) in ekus:
        if _has_full_name_constraints(crypto_cert):
            return serverauth_constants.CertificateType.INTERNAL_CONSTRAINED_TLS_CA
        else:
            return serverauth_constants.CertificateType.INTERNAL_UNCONSTRAINED_TLS_CA
    else:
        return serverauth_constants.CertificateType.NON_TLS_CA


def _is_ca(crypto_cert: x509.Certificate) -> bool:
    try:
        ext = crypto_cert.extensions.get_extension_for_oid(x509.OID_BASIC_CONSTRAINTS)

        return ext.value.ca
    except x509.ExtensionNotFound:
        return False


def _is_ocsp_responder(crypto_cert: x509.Certificate):
    try:
        ext = crypto_cert.extensions.get_extension_for_oid(x509.OID_EXTENDED_KEY_USAGE)

        return x509.OID_OCSP_SIGNING in ext.value
    except x509.ExtensionNotFound:
        return False


def _is_precert(crypto_cert: x509.Certificate):
    try:
        _ = crypto_cert.extensions.get_extension_for_class(x509.PrecertPoison)

        return True
    except x509.ExtensionNotFound:
        return False


def _determine_subscriber_certificate_type(crypto_cert: x509.Certificate):
    is_precert = _is_precert(crypto_cert)

    try:
        ext = crypto_cert.extensions.get_extension_for_oid(x509.OID_CERTIFICATE_POLICIES)

        policy_oids = {pi.policy_identifier.dotted_string for pi in ext.value}

        if str(serverauth_constants.ID_POLICY_EV) in policy_oids:
            return (serverauth_constants.CertificateType.EV_PRE_CERTIFICATE if is_precert
                    else serverauth_constants.CertificateType.EV_FINAL_CERTIFICATE)
        elif str(serverauth_constants.ID_POLICY_IV) in policy_oids:
            return (serverauth_constants.CertificateType.IV_PRE_CERTIFICATE if is_precert
                    else serverauth_constants.CertificateType.IV_FINAL_CERTIFICATE)
        elif str(serverauth_constants.ID_POLICY_OV) in policy_oids:
            return (serverauth_constants.CertificateType.OV_PRE_CERTIFICATE if is_precert
                    else serverauth_constants.CertificateType.OV_FINAL_CERTIFICATE)
        else:
            return (serverauth_constants.CertificateType.DV_PRE_CERTIFICATE if is_precert
                    else serverauth_constants.CertificateType.DV_FINAL_CERTIFICATE)
    except x509.ExtensionNotFound:
        # at a loss, guess DV?
        return (serverauth_constants.CertificateType.DV_PRE_CERTIFICATE if is_precert
                else serverauth_constants.CertificateType.DV_FINAL_CERTIFICATE)


def determine_certificate_type(cert: certificate.RFC5280Certificate) -> serverauth_constants.CertificateType:
    if cert.is_self_issued:
        return serverauth_constants.CertificateType.ROOT_CA

    crypto_cert = cert.cryptography_object

    if _is_ca(crypto_cert):
        return _get_intermediate_ca_type(crypto_cert)
    else:
        if _is_ocsp_responder(crypto_cert):
            return serverauth_constants.CertificateType.OCSP_RESPONDER
        else:
            return _determine_subscriber_certificate_type(crypto_cert)


def create_decoding_validators():
    return pkilint.pkix.certificate.create_decoding_validators(
        cabf.NAME_ATTRIBUTE_MAPPINGS,
        cabf.EXTENSION_MAPPINGS,
        [certificate.create_other_name_decoder(OTHER_NAME_MAPPINGS),
         certificate.create_qc_statements_decoder(etsi.ETSI_QC_STATEMENTS_MAPPINGS)]
    )


def create_top_level_certificate_validators(certificate_type: serverauth_constants.CertificateType):
    validators = [
        serverauth_key.ServerauthAllowedSignatureAlgorithmEncodingValidator(
            path='certificate.tbsCertificate.signature'
        ),
        serverauth_key.ServerauthAllowedSignatureAlgorithmEncodingValidator(
            path='certificate.signatureValue'
        ),
        cabf_extension.CabfExtensionsPresenceValidator(),
    ]

    if certificate_type == serverauth_constants.CertificateType.ROOT_CA:
        validators.append(serverauth_root.RootExtensionAllowanceValidator())
    elif certificate_type in serverauth_constants.CROSS_CA_TYPES:
        validators.append(serverauth_cross_ca.CrossCertificateExtensionAllowanceValidator(certificate_type))
    elif certificate_type in serverauth_constants.INTERMEDIATE_CERTIFICATE_TYPES:
        validators.append(serverauth_ca.CaCertificateExtensionAllowanceValidator(certificate_type))
    elif certificate_type in serverauth_constants.SUBSCRIBER_CERTIFICATE_TYPES:
        validators.append(serverauth_subscriber.SubscriberExtensionAllowanceValidator(certificate_type))
    elif certificate_type == serverauth_constants.CertificateType.OCSP_RESPONDER:
        validators.append(serverauth_ocsp.OcspExtensionAllowanceValidator())
    else:
        raise ValueError(f'Unsupported certificate type: {certificate_type}')

    return validators


def create_spki_validator_container():
    return validation.ValidatorContainer(validators=[
        serverauth_key.ServerauthAllowedPublicKeyAlgorithmEncodingValidator(
            path='certificate.tbsCertificate.subjectPublicKeyInfo.algorithm'
        ),
        cabf_key.RsaKeyValidator(),
        cabf_key.EcdsaKeyValidator(),
    ],
        path='certificate.tbsCertificate.subjectPublicKeyInfo')


def create_subject_name_validators() -> List[validation.Validator]:
    return [
        serverauth_name.AttributeOrderEncodingValidator(),
        serverauth_name.AttributeValueDirectoryStringValidator(),
        serverauth_name.X520NameAttributeValueLengthValidator(),
        serverauth_name.DomainComponentAttributeValueLengthValidator(),
        serverauth_name.ValidJurisdictionCountryValidator(),
        cabf_name.ValidCountryValidator(),
        serverauth_name.ValidBusinessCategoryValidator(),
        cabf_name.OrganizationIdentifierAttributeValidator(relax_stateprovince_syntax=False),
        serverauth_name.ServerauthRelativeDistinguishedNameContainsOneElementValidator(),
    ]


def create_ca_name_validator_container(certificate_type: serverauth_constants.CertificateType):
    validators = [
        serverauth_ca.CaRequiredSubjectAttributesValidator(certificate_type),
        serverauth_name.ServerauthDuplicateAttributeTypeValidator(certificate_type),
    ]
    validators.extend(create_subject_name_validators())

    if certificate_type == serverauth_constants.CertificateType.ROOT_CA:
        validators.append(serverauth_root.RootSubjectIssuerIdenticalEncodingValidator())

    return certificate.create_subject_validator_container(validators)


def create_subscriber_name_validator_container(certificate_type: serverauth_constants.CertificateType):
    validators = create_subject_name_validators()
    validators.extend([
        serverauth_subscriber.SubscriberCommonNameValidator(),
        serverauth_name.ServerauthDuplicateAttributeTypeValidator(certificate_type),
    ])

    if certificate_type in serverauth_constants.EV_CERTIFICATE_TYPES:
        validators.extend([
            serverauth_subscriber.EvSubscriberAttributeAllowanceValidator(),
            serverauth_subscriber.EvSubscriberJurisdictionPresenceValidator(),
            serverauth_name.OrganizationIdentifierConsistentSubjectAndExtensionValidator()
        ])
    elif certificate_type in serverauth_constants.IV_CERTIFICATE_TYPES:
        validators.append(serverauth_subscriber.IvSubscriberAttributeAllowanceValidator())
    elif certificate_type in serverauth_constants.OV_CERTIFICATE_TYPES:
        validators.append(serverauth_subscriber.OvSubscriberAttributeAllowanceValidator())
    elif certificate_type in serverauth_constants.DV_CERTIFICATE_TYPES:
        validators.append(serverauth_subscriber.DvSubcriberAttributeAllowanceValidator())
    else:
        raise ValueError(f'Unsupported certificate type: {certificate_type}')

    if certificate_type in serverauth_constants.IDENTITY_CERTIFICATE_TYPES:
        validators.append(serverauth_subscriber.IdentityCertificateStateProvinceAndLocalityPresenceValidator())

    return certificate.create_subject_validator_container(validators)


def create_extension_validators() -> List[validation.Validator]:
    return [
        serverauth_extension.CrlDpDistributionPointCountValidator(),
        serverauth_extension.CrlDpDistributionPointNameValidator(),
        serverauth_extension.CrlDpDistributionPointValidator(),
        serverauth_extension.AuthorityInformationAccessHttpUriLocationValidator(),
        serverauth_extension.AuthorityInformationAccessUniqueLocationValidator(),
        cabf_extension.CabfAuthorityKeyIdentifierValidator(),
        cabf_name.GeneralNameDnsNameInternalDomainNameValidator(allow_onion_tld=True),
        cabf_name.GeneralNameRfc822NameInternalDomainNameValidator(),
        cabf_name.GeneralNameUriInternalDomainNameValidator(),
        cabf_name.UriInternalDomainNameValidator(pdu_class=rfc5280.CPSuri),
        cabf_name.GeneralNameInternalIpAddressValidator(),
        serverauth_extension.CertificatePolicyQualifierValidator(),
        cabf_extension.CpsUriHttpValidator(),
        serverauth_name.DnsNameLdhLabelSyntaxValidator(),
        serverauth_name.TorVersion3DomainNameValidator(),
    ]


def create_ca_extension_validator_container(certificate_type: serverauth_constants.CertificateType):
    validators = create_extension_validators()

    validators.extend([
        serverauth_ca.CaCertificateExtensionCriticalityValidator(),
        serverauth_ca.NameConstraintsBaseTypeValidator(),
        serverauth_ca.CaCertificatePoliciesValidator(certificate_type),
        serverauth_ca.CaCertificateAuthorityInformationAccessAccessMethodPresenceValidator(),
        cabf_ca.CaKeyUsageValidator(),
        cabf_ca.CaBasicConstraintsValidator(),
    ])

    if certificate_type == serverauth_constants.CertificateType.ROOT_CA:
        validators.extend([
            serverauth_root.RootAkiSkiEqualityValidator(),
            serverauth_root.RootBasicConstraintsValidator(),
        ])
    elif certificate_type in serverauth_constants.CROSS_CA_TYPES:
        validators.append(serverauth_cross_ca.CrossCertificateAllowedEkuValidator(certificate_type))
    elif certificate_type == serverauth_constants.CertificateType.NON_TLS_CA:
        validators.append(serverauth_ca.NonTlsCaCertificateAllowedEkuValidator())
    elif certificate_type == serverauth_constants.CertificateType.PRECERT_SIGNING_CA:
        validators.append(serverauth_ca.PrecertSigningCaCertificateAllowedEkuValidator())
    elif certificate_type in serverauth_constants.TLS_CA_TYPES:
        validators.append(serverauth_ca.TlsCaCertificateAllowedEkuValidator())

    if certificate_type in {serverauth_constants.CertificateType.EXTERNAL_UNCONSTRAINED_EV_TLS_CA,
                            serverauth_constants.CertificateType.EXTERNAL_CONSTRAINED_EV_TLS_CA}:
        validators.append(serverauth_extension.EvCpsUriPresenceValidator())

    if certificate_type in serverauth_constants.CONSTRAINED_TLS_CA_TYPES:
        validators.append(serverauth_ca.TlsCaTechnicallyConstrainedValidator())

    return certificate.create_extensions_validator_container(validators)


def create_subscriber_extension_validator_container(certificate_type: serverauth_constants.CertificateType):
    validators = create_extension_validators()

    validators.extend([
        serverauth_subscriber.SubscriberEkuAllowanceValidator(),
        serverauth_subscriber.CABFOrganizationIdentifierExtensionValidator(),
        serverauth_subscriber.SubscriberExtensionCriticalityValidator(),
        serverauth_subscriber.SubscriberAuthorityInformationAccessAccessMethodPresenceValidator(),
        serverauth_subscriber.SubscriberKeyUsageValidator(),
        serverauth_subscriber.SubscriberBasicConstraintsValidator(),
        serverauth_subscriber.SubscriberPoliciesValidator(certificate_type),
    ])

    if certificate_type in serverauth_constants.EV_CERTIFICATE_TYPES:
        validators.extend([
            serverauth_subscriber.EvSanGeneralNameTypeValidator(),
            serverauth_extension.EvCpsUriPresenceValidator(),
            serverauth_subscriber.EvWildcardAllowanceValidator(),
        ])
    else:
        validators.append(serverauth_subscriber.SubscriberSanGeneralNameTypeValidator())

    return certificate.create_extensions_validator_container(validators)


def create_ocsp_extension_validator_container():
    validators = create_extension_validators()

    validators.extend([
        serverauth_ocsp.OcspEkuAllowanceValidator(),
        serverauth_ocsp.OcspAuthorityInformationAccessAccessMethodPresenceValidator(),
        serverauth_ocsp.OcspResponderKeyUsageValidator(),
        serverauth_ocsp.OcspBasicConstraintsValidator(),
    ])

    return certificate.create_extensions_validator_container(validators)


def create_validity_validator_container(certificate_type):
    validators = []

    if certificate_type == serverauth_constants.CertificateType.ROOT_CA:
        validators.append(serverauth_root.RootValidityPeriodValidator())
    elif certificate_type in serverauth_constants.SUBSCRIBER_CERTIFICATE_TYPES:
        validators.append(serverauth_subscriber.SubscriberValidityPeriodValidator())

    return certificate.create_validity_validator_container(validators)


def create_root_ca_validators():
    return [
        create_validity_validator_container(serverauth_constants.CertificateType.ROOT_CA),
        create_spki_validator_container(),
        create_ca_name_validator_container(serverauth_constants.CertificateType.ROOT_CA),
        create_ca_extension_validator_container(serverauth_constants.CertificateType.ROOT_CA),
    ] + create_top_level_certificate_validators(serverauth_constants.CertificateType.ROOT_CA)


def create_intermediate_ca_validators(certificate_type: serverauth_constants.CertificateType):
    return [
        create_validity_validator_container(certificate_type),
        create_spki_validator_container(),
        create_ca_name_validator_container(certificate_type),
        create_ca_extension_validator_container(certificate_type),
    ] + create_top_level_certificate_validators(certificate_type)


def create_subscriber_validators(certificate_type: serverauth_constants.CertificateType):
    return [
        create_validity_validator_container(certificate_type),
        create_spki_validator_container(),
        create_subscriber_name_validator_container(certificate_type),
        create_subscriber_extension_validator_container(certificate_type),
    ] + create_top_level_certificate_validators(certificate_type)


def create_ocsp_responder_validators():
    return [
        create_validity_validator_container(serverauth_constants.CertificateType.OCSP_RESPONDER),
        create_spki_validator_container(),
        create_ca_name_validator_container(serverauth_constants.CertificateType.OCSP_RESPONDER),
        create_ocsp_extension_validator_container(),
    ] + create_top_level_certificate_validators(serverauth_constants.CertificateType.OCSP_RESPONDER)


def create_validators(certificate_type: serverauth_constants.CertificateType):
    if certificate_type == serverauth_constants.CertificateType.ROOT_CA:
        return create_root_ca_validators()
    elif certificate_type in serverauth_constants.INTERMEDIATE_CERTIFICATE_TYPES:
        return create_intermediate_ca_validators(certificate_type)
    elif certificate_type in serverauth_constants.SUBSCRIBER_CERTIFICATE_TYPES:
        return create_subscriber_validators(certificate_type)
    elif certificate_type == serverauth_constants.CertificateType.OCSP_RESPONDER:
        return create_ocsp_responder_validators()
    else:
        raise ValueError(f'Unsupported certificate type: {certificate_type}')


def create_serverauth_finding_filters(certificate_type: serverauth_constants.CertificateType):
    filters = [
        serverauth_finding_filter.NameConstraintsCriticalityFilter(),
        serverauth_finding_filter.PolicyQualifierPresentFilter(),
        serverauth_finding_filter.DnsNameGeneralNamePreferredNameSyntaxFilter(),
    ]

    if certificate_type in serverauth_constants.SUBSCRIBER_CERTIFICATE_TYPES:
        filters += [
            serverauth_finding_filter.EndEntitySubjectKeyIdentifierMissingFilter(),
        ]

    return filters



from pyasn1_alt_modules import rfc5280

from pkilint import validation


class SubjectEmailAddressInSanValidator(validation.Validator):
    VALIDATION_SUBJECT_EMAIL_NOT_IN_SAN = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'pkix.subject_email_address_not_in_san'
    )

    def __init__(self):
        super().__init__(
            pdu_class=rfc5280.EmailAddress,
            validations=[self.VALIDATION_SUBJECT_EMAIL_NOT_IN_SAN]
        )

    def validate(self, node):
        san_ext_idx = node.document.get_extension_by_oid(
            rfc5280.id_ce_subjectAltName
        )

        if san_ext_idx is None:
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_SUBJECT_EMAIL_NOT_IN_SAN,
                'Certificate does not have SAN extension'
            )

        ext, _ = san_ext_idx

        email_address = str(node.pdu)

        for gn in ext.navigate('extnValue.subjectAltName').children.values():
            value_node = gn.children.get('rfc822Name')
            if value_node is None:
                continue
            if str(value_node.pdu) == email_address:
                return

        raise validation.ValidationFindingEncountered(
            self.VALIDATION_SUBJECT_EMAIL_NOT_IN_SAN,
            f'Subject DN e-mail address "{email_address}" not found in SAN'
        )



import datetime
import enum
from typing import NamedTuple, List

from pyasn1_alt_modules import rfc6962

from pkilint import validation


class HashAlgorithm(enum.IntEnum):
    NONE = 0
    MD5 = 1
    SHA1 = 2
    SHA224 = 3
    SHA256 = 4
    SHA384 = 5
    SHA512 = 6


class SignatureAlgorithm(enum.IntEnum):
    ANONYMOUS = 0
    RSA = 1
    DSA = 2
    ECDSA = 3


_UNIX_EPOCH_DATETIME = datetime.datetime(1970, 1, 1, 0, 0, 0, tzinfo=datetime.timezone.utc)


class SignedCertificateTimestamp(NamedTuple):
    sct_version: int
    log_id: bytes
    timestamp_msec: int
    extensions: bytes
    hash_alg: HashAlgorithm
    sig_alg: SignatureAlgorithm
    signature: bytes
    raw: bytes

    @property
    def timestamp_datetime(self):
        return _UNIX_EPOCH_DATETIME + datetime.timedelta(milliseconds=self.timestamp_msec)


# Thank you JHA
# https://letsencrypt.org/2018/04/04/sct-encoding.html
def _decode_sct(octets: bytes, offset: int):
    initial_offset = offset

    length = int.from_bytes(octets[offset:offset + 2], 'big')
    offset += 2

    version = octets[offset]
    offset += 1

    log_id = octets[offset:offset + 32]
    offset += 32

    timestamp = int.from_bytes(octets[offset: offset + 8], 'big')
    offset += 8

    extensions_length = int.from_bytes(octets[offset: offset + 2], 'big')
    offset += 2

    extensions = b'' if extensions_length == 0 else octets[offset:offset + extensions_length]
    offset += extensions_length

    raw = octets[initial_offset:offset]

    hash_alg = HashAlgorithm(octets[offset])
    offset += 1

    sig_alg = SignatureAlgorithm(octets[offset])
    offset += 1

    signature_length = int.from_bytes(octets[offset:offset + 2], 'big')
    offset += 2

    signature_octets = octets[offset:offset + signature_length]
    offset += signature_length

    return offset, SignedCertificateTimestamp(
        version, log_id, timestamp, extensions, hash_alg, sig_alg, signature_octets, raw)


def _decode(instance) -> List[SignedCertificateTimestamp]:
    octets = instance.asOctets()

    expected_length = int.from_bytes(octets[0:2], 'big')
    offset = 2

    actual_length = len(octets) - 2

    if actual_length != expected_length:
        raise ValueError('Invalid SCT list encoding: '
                         f'expected length: {expected_length}, actual length: {actual_length}')

    scts = []
    while offset < actual_length:
        offset, sct = _decode_sct(octets, offset)

        scts.append(sct)

    return scts


class SctListExtensionDecodingValidator(validation.Validator):
    VALIDATION_SCT_EXTENSION_INVALID_ENCODING = validation.ValidationFinding(
        validation.ValidationFindingSeverity.FATAL,
        'pkix.sct_list_extension_invalid_encoding'
    )

    def __init__(self, append_decoded=True):
        super().__init__(validations=self.VALIDATION_SCT_EXTENSION_INVALID_ENCODING,
                         pdu_class=rfc6962.SignedCertificateTimestampList)

        self._append_decoded = append_decoded

    def validate(self, node):
        try:
            decoded = _decode(node.pdu)

            if self._append_decoded:
                setattr(node.pdu, 'decoded', decoded)
        except ValueError as e:
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_SCT_EXTENSION_INVALID_ENCODING,
                str(e)
            )


class SctListElementCountValidator(validation.Validator):
    VALIDATION_SCT_LIST_EMPTY = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'pkix.sct_list_empty'
    )

    def __init__(self):
        super().__init__(validations=self.VALIDATION_SCT_LIST_EMPTY, pdu_class=rfc6962.SignedCertificateTimestampList)

    def validate(self, node):
        try:
            decoded = getattr(node.pdu, 'decoded')
        except AttributeError:
            return

        if len(decoded) == 0:
            raise validation.ValidationFindingEncountered(self.VALIDATION_SCT_LIST_EMPTY)



from pkilint import document, validation, report


class DummyValidator1(validation.Validator):
    pass


class DummyValidator2(validation.Validator):
    pass


_DUMMY_NODE_1 = document.PDUNode(None, 'foo', None, None)
_DUMMY_NODE_2 = document.PDUNode(None, 'bar', None, None)


_RESULTS = [
    validation.ValidationResult(DummyValidator1(), _DUMMY_NODE_1, []),
    validation.ValidationResult(DummyValidator2(), _DUMMY_NODE_2, [
        validation.ValidationFindingDescription(
            validation.ValidationFinding(severity=validation.ValidationFindingSeverity.INFO, code='info_finding'),
            None
        ),
        validation.ValidationFindingDescription(
            validation.ValidationFinding(severity=validation.ValidationFindingSeverity.ERROR, code='error_finding'),
            'The error message'
        )
    ])
]


def test_plaintext_all():
    gen = report.ReportGeneratorPlaintext(_RESULTS, None)

    assert gen.generate() == 'DummyValidator1 @ foo\nDummyValidator2 @ bar\n    info_finding (INFO)\n    error_finding (ERROR): The error message\n'


def test_plaintext_info():
    gen = report.ReportGeneratorPlaintext(_RESULTS, validation.ValidationFindingSeverity.INFO)

    assert gen.generate() == 'DummyValidator2 @ bar\n    info_finding (INFO)\n    error_finding (ERROR): The error message\n'


def test_plaintext_warning():
    gen = report.ReportGeneratorPlaintext(_RESULTS, validation.ValidationFindingSeverity.WARNING)

    assert gen.generate() == 'DummyValidator2 @ bar\n    error_finding (ERROR): The error message\n'


def test_csv_all():
    gen = report.ReportGeneratorCsv(_RESULTS, None)

    assert gen.generate() == 'node_path,validator,severity,code,message\r\nbar,DummyValidator2,INFO,info_finding,\r\nbar,DummyValidator2,ERROR,error_finding,The error message\r\n'


def test_csv_info():
    gen = report.ReportGeneratorCsv(_RESULTS, validation.ValidationFindingSeverity.INFO)

    assert gen.generate() == 'node_path,validator,severity,code,message\r\nbar,DummyValidator2,INFO,info_finding,\r\nbar,DummyValidator2,ERROR,error_finding,The error message\r\n'


def test_csv_warning():
    gen = report.ReportGeneratorCsv(_RESULTS, validation.ValidationFindingSeverity.WARNING)

    assert gen.generate() == 'node_path,validator,severity,code,message\r\nbar,DummyValidator2,ERROR,error_finding,The error message\r\n'


def test_json_all():
    gen = report.ReportGeneratorJson(_RESULTS, None)

    assert gen.generate() == '{"results": [{"node_path": "foo", "validator": "DummyValidator1", "finding_descriptions": []}, {"node_path": "bar", "validator": "DummyValidator2", "finding_descriptions": [{"severity": "INFO", "code": "info_finding", "message": null}, {"severity": "ERROR", "code": "error_finding", "message": "The error message"}]}]}'


def test_json_info():
    gen = report.ReportGeneratorJson(_RESULTS, validation.ValidationFindingSeverity.INFO)

    assert gen.generate() == '{"results": [{"node_path": "bar", "validator": "DummyValidator2", "finding_descriptions": [{"severity": "INFO", "code": "info_finding", "message": null}, {"severity": "ERROR", "code": "error_finding", "message": "The error message"}]}]}'


def test_json_warning():
    gen = report.ReportGeneratorJson(_RESULTS, validation.ValidationFindingSeverity.WARNING)

    assert gen.generate() == '{"results": [{"node_path": "bar", "validator": "DummyValidator2", "finding_descriptions": [{"severity": "ERROR", "code": "error_finding", "message": "The error message"}]}]}'




from pyasn1_alt_modules import rfc5280, rfc6960, rfc6962

import pkilint.common
from pkilint import validation, common
from pkilint.itu import bitstring
from pkilint.pkix import Rfc2119Word


_CODE_CLASSIFIER = 'cabf.serverauth.ocsp_responder'


class OcspResponderKeyUsageValidator(validation.Validator):
    """Validates that the content of the key usage extension conforms with BR 7.1.2.8.7."""

    VALIDATION_DIGSIG_MISSING = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        _CODE_CLASSIFIER + '.digitalsignature_bit_missing'
    )

    VALIDATION_PROHIBITED_KU_PRESENT = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        _CODE_CLASSIFIER + '.prohibited_ku_present'
    )

    _PROHIBITED_KUS = {
                          str(n) for n in rfc5280.KeyUsage.namedValues
                      } - {'digitalSignature'}

    def __init__(self):
        super().__init__(validations=[self.VALIDATION_DIGSIG_MISSING, self.VALIDATION_PROHIBITED_KU_PRESENT],
                         pdu_class=rfc5280.KeyUsage)

    def validate(self, node):
        if not bitstring.has_named_bit(node, 'digitalSignature'):
            raise validation.ValidationFindingEncountered(self.VALIDATION_DIGSIG_MISSING)

        prohibited_kus_asserted = sorted((k for k in self._PROHIBITED_KUS if bitstring.has_named_bit(node, k)))

        if any(prohibited_kus_asserted):
            prohibited_kus_str = ', '.join(prohibited_kus_asserted)

            raise validation.ValidationFindingEncountered(
                self.VALIDATION_PROHIBITED_KU_PRESENT, f'Prohibited KUs present: {prohibited_kus_str}'
            )


class OcspAuthorityInformationAccessAccessMethodPresenceValidator(
        common.AuthorityInformationAccessAccessMethodPresenceValidator):
    """Validates that the content of the AIA extension conforms with BR 7.1.2.8.3."""

    _ACCESS_METHOD_ALLOWANCES = {
        rfc5280.id_ad_ocsp: Rfc2119Word.SHOULD_NOT,
    }

    def __init__(self):
        super().__init__(self._ACCESS_METHOD_ALLOWANCES, _CODE_CLASSIFIER, Rfc2119Word.MUST_NOT)


class OcspExtensionAllowanceValidator(pkilint.common.ExtensionIdentifierAllowanceValidator):
    """Validates that the included extensions conform with BR 7.1.2.8.2."""

    _EXTENSION_ALLOWANCES = {
        rfc5280.id_ce_authorityKeyIdentifier: Rfc2119Word.MUST,
        rfc5280.id_ce_extKeyUsage: Rfc2119Word.MUST,
        rfc6960.id_pkix_ocsp_nocheck: Rfc2119Word.MUST,
        rfc5280.id_ce_keyUsage: Rfc2119Word.MUST,
        rfc5280.id_ce_basicConstraints: Rfc2119Word.MAY,
        rfc5280.id_ce_nameConstraints: Rfc2119Word.MUST_NOT,
        rfc5280.id_ce_subjectAltName: Rfc2119Word.MUST_NOT,
        rfc5280.id_ce_subjectKeyIdentifier: Rfc2119Word.SHOULD,
        rfc5280.id_pe_authorityInfoAccess: Rfc2119Word.SHOULD_NOT,
        rfc5280.id_ce_certificatePolicies: Rfc2119Word.MUST_NOT,
        rfc5280.id_ce_cRLDistributionPoints: Rfc2119Word.MUST_NOT,
        rfc6962.id_ce_embeddedSCT: Rfc2119Word.MAY,
    }

    def __init__(self):
        super().__init__(self._EXTENSION_ALLOWANCES, _CODE_CLASSIFIER, Rfc2119Word.SHOULD_NOT)


class OcspEkuAllowanceValidator(pkilint.common.ExtendedKeyUsageAllowanceValidator):
    """Validates that the extended key usage value conforms to BR 7.1.2.8.5."""

    _EKU_ALLOWANCES = {
        rfc5280.id_kp_OCSPSigning: Rfc2119Word.MUST,
    }

    def __init__(self):
        super().__init__(self._EKU_ALLOWANCES, _CODE_CLASSIFIER, Rfc2119Word.MUST_NOT)


class OcspBasicConstraintsValidator(validation.Validator):
    """Validates that the basic constraints extension value conforms to BR 7.1.2.8.4."""

    VALIDATION_CA_BIT_SET = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        _CODE_CLASSIFIER + '.basic_constraints_ca_bit_set'
    )

    def __init__(self):
        super().__init__(validations=self.VALIDATION_CA_BIT_SET, pdu_class=rfc5280.BasicConstraints)

    def validate(self, node):
        if bool(node.children['cA'].pdu):
            raise validation.ValidationFindingEncountered(self.VALIDATION_CA_BIT_SET)



#!/usr/bin/env python

import argparse
import sys

from pyasn1_alt_modules import rfc5280

from pkilint import loader, document, util, validation, pkix, report
from pkilint.pkix import certificate, name, extension, algorithm
from pkilint.pkix.certificate import certificate_extension, certificate_key


def create_decoder_validation_container():
    decoders = [
        pkix.create_attribute_decoder(name.ATTRIBUTE_TYPE_MAPPINGS),
        pkix.create_extension_decoder(extension.EXTENSION_MAPPINGS),
        pkix.create_signature_algorithm_identifier_decoder(
            algorithm.SIGNATURE_ALGORITHM_IDENTIFIER_MAPPINGS,
            path='certificate.tbsCertificate.signature'
        ),
        certificate.create_spki_decoder(
            certificate_key.SUBJECT_PUBLIC_KEY_ALGORITHM_IDENTIFIER_MAPPINGS,
            certificate_key.SUBJECT_KEY_PARAMETER_ALGORITHM_IDENTIFIER_MAPPINGS
        ),
    ]

    return validation.ValidatorContainer(
        validators=decoders, path='certificate'
    )


def create_issuer_validation_container():
    validators = [
        name.IssuerSubjectNameBinaryEqualValidator(
            path='certificate.tbsCertificate.subject',
            subject_document_issuer_dn_path='subject:certificate.tbsCertificate.issuer'
        ),
        extension.IssuerSubjectKeyIdentifierBinaryEqualValidator(
            subject_auth_key_id_retriever=(
                lambda n: document.get_document_by_name(n, 'subject').get_extension_by_oid(
                    rfc5280.id_ce_authorityKeyIdentifier
                )
            )
        ),
        certificate_extension.IssuerSubjectPolicyChainValidator(),
    ]

    return validation.ValidatorContainer(
        validators=validators
    )


def create_subject_validation_container():
    validators = [
        certificate_key.SubjectSignatureVerificationValidator(
            tbs_node_retriever=lambda n: n.navigate('^.tbsCertificate'),
            path='certificate.signature'
        )
    ]

    return validation.ValidatorContainer(
        validators=validators
    )


def main(cli_args=None) -> int:
    parser = argparse.ArgumentParser('RFC 5280 Issuer/Subject Certificate Chain Linter')

    subparsers = parser.add_subparsers(dest='command', required=True)
    subparsers.add_parser('validations', help='Output the set of validations which this linter performs')

    lint_parser = subparsers.add_parser('lint', help='Lint the specified issuer and subject certificates')
    util.add_standard_args(lint_parser)

    lint_parser.add_argument(dest='issuer', type=argparse.FileType('rb'),
                             help='The issuer certificate to lint'
                             )
    lint_parser.add_argument(dest='subject', type=argparse.FileType('rb'),
                             help='The subject certificate to lint'
                             )

    args = parser.parse_args(cli_args)

    decoding_validation_container = create_decoder_validation_container()
    issuer_validation_container = create_issuer_validation_container()
    subject_validation_container = create_subject_validation_container()

    if args.command == 'validations':
        print(report.report_included_validations(decoding_validation_container, issuer_validation_container,
                                                 subject_validation_container))

        return 0
    else:
        doc_collection = {}

        issuer = loader.load_certificate(args.issuer, args.issuer.name, 'issuer',
                                         doc_collection
                                         )
        doc_collection['issuer'] = issuer

        subject = loader.load_certificate(args.subject, args.subject.name, 'subject',
                                          doc_collection
                                          )
        doc_collection['subject'] = subject

        results = decoding_validation_container.validate(issuer.root)
        results += decoding_validation_container.validate(subject.root)
        results += issuer_validation_container.validate(issuer.root)
        results += subject_validation_container.validate(subject.root)

        print(args.format(results, args.severity))

        return report.get_findings_count(results, args.severity)


if __name__ == '__main__':
    sys.exit(main())



from pkilint import finding_filter
from pkilint.pkix import general_name
from pkilint.pkix.certificate import certificate_extension


class NameConstraintsCriticalityFilter(finding_filter.ValidationFindingFilter):
    def __init__(self):
        super().__init__(certificate_extension.NameConstraintsCriticalityValidator.VALIDATION_NC_NOT_CRITICAL)


class DnsNameGeneralNamePreferredNameSyntaxFilter(finding_filter.ValidationFindingFilter):
    def __init__(self):
        super().__init__(general_name.GeneralNameDnsNameSyntaxValidator.VALIDATION_NOT_PREFERRED_NAME_SYNTAX)


class EndEntitySubjectKeyIdentifierMissingFilter(finding_filter.ValidationFindingFilter):
    def __init__(self):
        super().__init__(certificate_extension.SubjectKeyIdentifierPresenceValidator.VALIDATION_EE_SKID_MISSING)
        
        
class PolicyQualifierPresentFilter(finding_filter.ValidationFindingFilter):
    def __init__(self):
        super().__init__(certificate_extension.CertificatePolicyQualifierValidator.VALIDATION_POLICY_HAS_QUALIFIER)



# Auto-generated by asn1ate v.0.6.0 from en_319_412_5.asn1 with manual fixups
# Taken from ETSI EN 319-412-5, Annex B


from pyasn1.type import univ, char, namedtype, constraint

MAX = float('inf')


def _OID(*components):
    output = []
    for x in tuple(components):
        if isinstance(x, univ.ObjectIdentifier):
            output.extend(list(x))
        else:
            output.append(int(x))

    return univ.ObjectIdentifier(output)


class CountryName(char.PrintableString):
    pass


CountryName.subtypeSpec = constraint.ValueSizeConstraint(2, 2)


class Iso4217CurrencyCode(univ.Choice):
    pass


Iso4217CurrencyCode.componentType = namedtype.NamedTypes(
    namedtype.NamedType('alphabetic', char.PrintableString().subtype(subtypeSpec=constraint.ValueSizeConstraint(3, 3))),
    namedtype.NamedType('numeric', univ.Integer().subtype(subtypeSpec=constraint.ValueRangeConstraint(1, 999)))
)


class MonetaryValue(univ.Sequence):
    pass


MonetaryValue.componentType = namedtype.NamedTypes(
    namedtype.NamedType('currency', Iso4217CurrencyCode()),
    namedtype.NamedType('amount', univ.Integer()),
    namedtype.NamedType('exponent', univ.Integer())
)


class PdsLocation(univ.Sequence):
    pass


PdsLocation.componentType = namedtype.NamedTypes(
    namedtype.NamedType('url', char.IA5String()),
    namedtype.NamedType('language', char.PrintableString().subtype(subtypeSpec=constraint.ValueSizeConstraint(2, 2)))
)


class PdsLocations(univ.SequenceOf):
    pass


PdsLocations.componentType = PdsLocation()
PdsLocations.subtypeSpec = constraint.ValueSizeConstraint(1, MAX)


class QcCClegislation(univ.SequenceOf):
    pass


QcCClegislation.componentType = CountryName()


class QcEuLimitValue(MonetaryValue):
    pass


class QcEuPDS(PdsLocations):
    pass


class QcEuRetentionPeriod(univ.Integer):
    pass


class QcType(univ.SequenceOf):
    pass


QcType.componentType = univ.ObjectIdentifier()

id_etsi_qcs = _OID(0, 4, 0, 1862, 1)

id_etsi_qcs_QcCClegislation = _OID(id_etsi_qcs, 7)

id_etsi_qcs_QcCompliance = _OID(id_etsi_qcs, 1)

id_etsi_qcs_QcLimitValue = _OID(id_etsi_qcs, 2)

id_etsi_qcs_QcPDS = _OID(id_etsi_qcs, 5)

id_etsi_qcs_QcRetentionPeriod = _OID(id_etsi_qcs, 3)

id_etsi_qcs_QcSSCD = _OID(id_etsi_qcs, 4)

id_etsi_qcs_QcType = _OID(id_etsi_qcs, 6)

id_etsi_qct_eseal = _OID(id_etsi_qcs_QcType, 2)

id_etsi_qct_esign = _OID(id_etsi_qcs_QcType, 1)

id_etsi_qct_web = _OID(id_etsi_qcs_QcType, 3)



from pyasn1.type.constraint import ValueRangeConstraint
from pyasn1_alt_modules import rfc5280

from pkilint import validation
from pkilint.pkix import extension


class CrlNumberPresenceValidator(extension.ExtensionPresenceValidator):
    def __init__(self):
        finding = validation.ValidationFinding(
            validation.ValidationFindingSeverity.ERROR,
            'pkix.crl_number_missing'
        )

        super().__init__(extension_oid=rfc5280.id_ce_cRLNumber,
                         validation=finding, pdu_class=rfc5280.CertificateList
                         )


class CrlNumberValueValidator(validation.ASN1ConstraintValidator):
    VALIDATION_FINDING_CRL_NUMBER_OOR = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'pkix.crl_number_out_of_range'
    )

    MAX_VALUE = (1 << 159) - 1

    def __init__(self):
        super().__init__(
            pdu_class=rfc5280.CRLNumber,
            validations=[
                self.VALIDATION_FINDING_CRL_NUMBER_OOR
            ],
            constraint=ValueRangeConstraint(0, self.MAX_VALUE)
        )


class AuthorityKeyIdentifierPresenceValidator(extension.ExtensionPresenceValidator):
    def __init__(self):
        finding = validation.ValidationFinding(
            validation.ValidationFindingSeverity.ERROR,
            'pkix.authority_key_identifier_missing'
        )

        super().__init__(extension_oid=rfc5280.id_ce_authorityKeyIdentifier,
                         validation=finding, pdu_class=rfc5280.CertificateList
                         )


class CrlNumberCriticalityValidator(extension.ExtensionCriticalityValidator):
    def __init__(self):
        finding = validation.ValidationFinding(
            validation.ValidationFindingSeverity.ERROR,
            'pkix.crl_number_extension_critical'
        )

        super().__init__(type_oid=rfc5280.id_ce_cRLNumber, is_critical=False,
                         validation=finding
                         )


class CrlReasonCodeValidator(validation.Validator):
    VALIDATION_UNSPECIFIED_REASON_CODE = validation.ValidationFinding(
        validation.ValidationFindingSeverity.WARNING,
        'pkix.crl_unspecified_crl_entry_reason_code'
    )

    def __init__(self):
        super().__init__(
            pdu_class=rfc5280.CRLReason,
            validations=[self.VALIDATION_UNSPECIFIED_REASON_CODE]
        )

    def validate(self, node):
        unspecified = rfc5280.CRLReason.namedValues['unspecified']

        if node.pdu == unspecified:
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_UNSPECIFIED_REASON_CODE
            )


class CrlReasonCodeCriticalityValidator(extension.ExtensionCriticalityValidator):
    def __init__(self):
        finding = validation.ValidationFinding(
            validation.ValidationFindingSeverity.ERROR,
            'pkix.crl_reason_code_extension_critical'
        )

        super().__init__(type_oid=rfc5280.id_ce_cRLReasons, is_critical=False,
                         validation=finding
                         )


class CrlReasonCodeAllowlistValidator(validation.Validator):
    VALIDATION_PROHIBITED_CRL_REASON_CODE = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'pkix.crl_prohibited_reason_code'
    )

    def __init__(self, *, allowed_reason_codes):
        self._allowed_reason_codes = allowed_reason_codes

        super().__init__(
            pdu_class=rfc5280.CRLReason,
            validations=[self.VALIDATION_PROHIBITED_CRL_REASON_CODE]
        )

    def validate(self, node):
        if node.pdu not in self._allowed_reason_codes:
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_PROHIBITED_CRL_REASON_CODE,
                f'Prohibited reason code "{node.pdu}"'
            )



from fastapi import HTTPException

from pkilint.cabf import smime
from pkilint.cabf.smime import smime_constants
from pkilint.pkix import certificate
from pkilint.rest import model


class CabfSmimeLinterGroup(model.LinterGroup):
    def __init__(self, linters):
        super().__init__(name='cabf-smime', linters=linters)

    def determine_linter(self, doc):
        v_g = smime.determine_validation_level_and_generation(doc)

        if v_g is None:
            raise HTTPException(status_code=422, detail='Could not determine certificate type')

        v, g = v_g

        name = f'{v}-{g}'

        return next((l for l in self.linters if l.name.casefold() == name.casefold()))


_V_G_PAIRS = []
for v in smime_constants.ValidationLevel:
    for g in smime_constants.Generation:
        _V_G_PAIRS.append((v, g))


def create_linter_group_instance():
    return CabfSmimeLinterGroup(
        [
            model.Linter(
                validator=certificate.create_pkix_certificate_validator_container(
                    smime.create_decoding_validators(),
                    smime.create_subscriber_validators(v, g)
                ),
                name=f'{v}-{g}'
            )
            for v, g in _V_G_PAIRS
        ]
    )



import pytest
from http import HTTPStatus
from importlib.metadata import version

from fastapi.testclient import TestClient

from pkilint.cabf.serverauth import serverauth_constants
from pkilint.cabf.smime import smime_constants
from pkilint.rest import app as web_app


@pytest.fixture()
def app():
    return web_app


@pytest.fixture()
def client(app):
    return TestClient(app)


def test_version(client):
    resp = client.get('/version')
    assert version('pkilint') == resp.json()['version']


_SMBR_SPONSORED_STRICT_PEM = '''-----BEGIN CERTIFICATE-----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-----END CERTIFICATE-----'''

_OV_FINAL_CLEAN_PEM = '''-----BEGIN CERTIFICATE-----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-----END CERTIFICATE-----'''


def test_groups(client):
    resp = client.get('/certificate')
    assert resp.status_code == HTTPStatus.OK

    j = resp.json()

    # SMIME and serverauth

    names = {lg['name'] for lg in j}
    assert names == {'cabf-serverauth', 'cabf-smime'}


def test_group_no_exist(client):
    resp = client.get('/certificate/foo')
    assert resp.status_code == HTTPStatus.NOT_FOUND


def test_group_different_case(client):
    # look at what staring at those tables in 7.1 for days has done to me
    resp = client.get('/certificate/CaBf-SeRveRauTh')

    assert resp.status_code == HTTPStatus.OK

    j = resp.json()

    names = {l['name'] for l in j['linters']}

    assert names == set(map(lambda c: c.to_option_str, serverauth_constants.CertificateType))


def test_smime_no_cert(client):
    resp = client.post('/certificate/cabf-smime', json={})
    assert resp.status_code == HTTPStatus.UNPROCESSABLE_ENTITY


def test_smime_bad_pem(client):
    resp = client.post('/certificate/cabf-smime', json={'pem': 'foo'})
    assert resp.status_code == HTTPStatus.UNPROCESSABLE_ENTITY


def test_smime_bad_base64(client):
    resp = client.post('/certificate/cabf-smime', json={'b64': 'foo'})
    assert resp.status_code == HTTPStatus.UNPROCESSABLE_ENTITY


def test_smime_detect(client):
    resp = client.post('/certificate/cabf-smime', json={'pem': _SMBR_SPONSORED_STRICT_PEM})

    j = resp.json()

    assert j['linter']['name'] == f'{smime_constants.ValidationLevel.SPONSORED}-{smime_constants.Generation.STRICT}'


def test_smime_detect_not_smime(client):
    resp = client.post('/certificate/cabf-smime', json={'pem': _OV_FINAL_CLEAN_PEM})
    assert resp.status_code == HTTPStatus.UNPROCESSABLE_ENTITY


def test_lint_smime_unknown_linter(client):
    resp = client.post('/certificate/cabf-smime/lint/FOOMASTER-BAR', json={'pem': _SMBR_SPONSORED_STRICT_PEM})
    assert resp.status_code == HTTPStatus.NOT_FOUND


def test_lint_smime(client):
    resp = client.post('/certificate/cabf-smime/SPONSORED-STRICT', json={'pem': _SMBR_SPONSORED_STRICT_PEM})
    assert resp.status_code == HTTPStatus.OK

    j = resp.json()

    assert len(j['results']) == 1

    finding_descriptions = j['results'][0]['finding_descriptions']
    assert len(finding_descriptions) == 1
    finding_description = finding_descriptions[0]

    assert finding_description['severity'] == 'INFO'
    assert finding_description['code'] == 'pkix.subject_key_identifier_method_1_identified'


def test_detect_and_lint_smime(client):
    resp = client.post('/certificate/cabf-smime', json={'pem': _SMBR_SPONSORED_STRICT_PEM})
    assert resp.status_code == HTTPStatus.OK

    j = resp.json()

    assert (
            j['linter']['name'] ==
            f'{smime_constants.ValidationLevel.SPONSORED.name}-{smime_constants.Generation.STRICT}'
    )

    assert len(j['results']) == 1

    finding_descriptions = j['results'][0]['finding_descriptions']
    assert len(finding_descriptions) == 1
    finding_description = finding_descriptions[0]

    assert finding_description['severity'] == 'INFO'
    assert finding_description['code'] == 'pkix.subject_key_identifier_method_1_identified'


def test_detect_and_lint_smime_with_tls(client):
    resp = client.post('/certificate/cabf-smime', json={'pem': _OV_FINAL_CLEAN_PEM})
    assert resp.status_code == HTTPStatus.UNPROCESSABLE_ENTITY


def test_serverauth_no_cert(client):
    resp = client.post('/certificate/cabf-serverauth', json={})
    assert resp.status_code == HTTPStatus.UNPROCESSABLE_ENTITY


def test_serverauth_bad_pem(client):
    resp = client.post('/certificate/cabf-serverauth', json={'pem': 'foo'})
    assert resp.status_code == HTTPStatus.UNPROCESSABLE_ENTITY


def test_serverauuth_bad_base64(client):
    resp = client.post('/certificate/cabf-serverauth', json={'b64': 'foo'})
    assert resp.status_code == HTTPStatus.UNPROCESSABLE_ENTITY


def test_serverauth_detect(client):
    resp = client.post('/certificate/cabf-serverauth', json={'pem': _OV_FINAL_CLEAN_PEM})

    j = resp.json()

    assert j['linter']['name'] == serverauth_constants.CertificateType.OV_FINAL_CERTIFICATE.to_option_str


def test_serverauth_detect_not_serverauth(client):
    resp = client.post('/certificate/cabf-serverauth', json={'pem': _SMBR_SPONSORED_STRICT_PEM})
    assert resp.status_code == HTTPStatus.OK

    j = resp.json()

    assert j['linter']['name'] == serverauth_constants.CertificateType.DV_FINAL_CERTIFICATE.to_option_str


def test_lint_serverauth_unknown_linter(client):
    resp = client.post('/certificate/cabf-serverauth/FOOMASTER-BAR', json={'pem': _OV_FINAL_CLEAN_PEM})
    assert resp.status_code == HTTPStatus.NOT_FOUND


def test_lint_serverauth(client):
    resp = client.post('/certificate/cabf-serverauth/OV-FINAL-CERTIFICATE', json={'pem': _OV_FINAL_CLEAN_PEM})
    assert resp.status_code == HTTPStatus.OK

    j = resp.json()

    assert len(j['results']) == 0


def test_detect_and_lint_serverauth(client):
    resp = client.post('/certificate/cabf-serverauth', json={'pem': _OV_FINAL_CLEAN_PEM})
    assert resp.status_code == HTTPStatus.OK

    j = resp.json()

    assert j['linter']['name'] == serverauth_constants.CertificateType.OV_FINAL_CERTIFICATE.to_option_str

    assert len(j['results']) == 0


def test_detect_and_lint_serverauth_with_smime(client):
    resp = client.post('/certificate/cabf-serverauth/', json={'pem': _SMBR_SPONSORED_STRICT_PEM})
    assert resp.status_code == HTTPStatus.OK

    j = resp.json()

    assert j['linter']['name'] == serverauth_constants.CertificateType.DV_FINAL_CERTIFICATE.to_option_str



import argparse
import functools

from cryptography.hazmat.primitives import hashes

from pkilint import validation, report
from pkilint.report import report_wrapper, REPORT_FORMATS


def _calculate_hash(octets, hash_algo):
    h = hashes.Hash(hash_algo)
    h.update(octets)

    return h.finalize()


def calculate_sha1_hash(octets):
    return _calculate_hash(octets, hashes.SHA1())


def argparse_enum_type_parser(enum_type):
    def _parse(value):
        value = value.upper()

        try:
            return enum_type[value]
        except KeyError:
            raise ValueError(value)

    return _parse


class SeverityThresholdAction(argparse.Action):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __call__(self, parser, namespace, values, option_string=None):
        if values == 'ALL':
            severity = None
        elif values not in [s.name for s in validation.ValidationFindingSeverity]:
            raise argparse.ArgumentError(self, f'Invalid severity value: "{values}"')
        else:
            severity = validation.ValidationFindingSeverity[values]

        setattr(namespace, self.dest, severity)


def add_severity_arg(parser):
    parser.add_argument('-s', '--severity',
                        type=str.upper,
                        default=validation.ValidationFindingSeverity.INFO,
                        help='The finding severity threshold; findings with a lesser severity will not be reported.',
                        action=SeverityThresholdAction,
                        choices=[s.name for s in validation.ValidationFindingSeverity] + ['ALL']
                        )


class ReportFormatAction(argparse.Action):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __call__(self, parser, namespace, values, option_string=None):
        report_generator_cls = REPORT_FORMATS[values]

        setattr(namespace, self.dest, functools.partial(report_wrapper, report_generator_cls))


def add_report_format_arg(parser):
    parser.add_argument('-f', '--format',
                        type=str.upper,
                        default=functools.partial(report_wrapper, report.ReportGeneratorPlaintext),
                        help='The format in which results will be output.',
                        choices=list(REPORT_FORMATS.keys()),
                        action=ReportFormatAction
                        )


def add_standard_args(parser):
    add_severity_arg(parser)
    add_report_format_arg(parser)



import enum
import math
import sys

from pyasn1.type.univ import ObjectIdentifier


BR_VERSION = '1.0.1'


CABF_SMIME_OID_ARC = ObjectIdentifier('2.23.140.1.5')


@enum.unique
class ValidationLevel(enum.IntEnum):
    MAILBOX = 1
    ORGANIZATION = 2
    SPONSORED = 4
    INDIVIDUAL = 8

    def __str__(self):
        return self.name


[
    setattr(
        sys.modules[__name__],
        f'CABF_SMIME_{v.name}_OID_ARC',
        CABF_SMIME_OID_ARC + (int(math.log2(v)) + 1,)
    )
    for v in ValidationLevel
]


@enum.unique
class Generation(enum.IntEnum):
    LEGACY = 1 << 8
    MULTIPURPOSE = 1 << 9
    STRICT = 1 << 10

    def __str__(self):
        return self.name


def _define_oids(validation_level):
    [
        setattr(
            sys.modules[__name__],
            f'CABF_SMIME_{validation_level.name}_{g.name}_OID',
            CABF_SMIME_OID_ARC + (
                int(math.log2(validation_level)) + 1,
                int(math.log2(g >> 8)) + 1,
            )
        )
        for g in Generation
    ]


[_define_oids(v) for v in ValidationLevel]


def get_policy_oid(validation_level, generation):
    return CABF_SMIME_OID_ARC + (
        int(math.log2(validation_level)) + 1,
        int(math.log2(generation >> 8)) + 1,
    )



import base64
from typing import List, Optional

from fastapi import HTTPException
from pyasn1.error import PyAsn1Error
from pydantic import BaseModel, Field, model_validator
from typing_extensions import Annotated

from pkilint import finding_filter, report, validation, loader, document


class Version(BaseModel):
    version: Annotated[str, Field(description='The version of pkilint that this server is using')]


_SEVERITY_DESCRIPTION = f'The severity of the finding ({", ".join(map(str, validation.ValidationFindingSeverity))})'


class FindingDescription(BaseModel):
    severity: Annotated[str, Field(description=_SEVERITY_DESCRIPTION)]
    code: Annotated[str, Field(description='The code that identifies the type of finding')]
    message: Annotated[Optional[str], Field(description='An optional message that provides further context for the '
                                                        'finding')]


class Result(BaseModel):
    validator: Annotated[str, Field(description='The class name of the validator which returned this result')]
    node_path: Annotated[str, Field(description='The path in the document (or set of documents) that was validated')]
    finding_descriptions: Annotated[
        List[FindingDescription],
        Field(description='The list of findings returned by the validator')
    ]


class LintResultList(BaseModel):
    results: Annotated[List[Result], Field(description='The list of results returned by the linter')]


class Linter(BaseModel):
    name: Annotated[str, Field(description='The name of the linter')]

    def __init__(self, validator, finding_filters=None, **kwargs):
        super().__init__(**kwargs)

        self._validator = validator
        self._finding_filters = finding_filters

    def lint(self, doc) -> LintResultList:
        results = self._validator.validate(doc.root)

        if self._finding_filters is not None:
            results, _ = finding_filter.filter_results(self._finding_filters, results)

        report_gen = report.ReportGeneratorJson(results, validation.ValidationFindingSeverity.INFO)
        json_str = report_gen.generate()

        return LintResultList.model_validate_json(json_str)


class LintResultListWithLinter(LintResultList):
    linter: Annotated[Linter, Field(description='The linter that was used for linting the specified document')]


class LinterGroup(BaseModel):
    name: Annotated[str, Field(description='The name of the linter group')]
    linters: Annotated[List[Linter], Field(description='The set of linters in this group')]

    def determine_linter(self, doc):
        pass

    def get_linter_by_name(self, name: str) -> Linter:
        try:
            return next((l for l in self.linters if l.name.casefold() == name.casefold()))
        except StopIteration:
            raise HTTPException(404, 'Linter with the specified name does not exist')


class DocumentModel(BaseModel):
    pem: Annotated[Optional[str], Field(description='A PEM-encoded ASN.1 document')] = None
    b64: Annotated[Optional[str], Field(description='A Base64-encoded DER representation of an ASN.1 document')] = None

    def _validate(self) -> 'DocumentModel':
        if self.pem and self.b64:
            raise ValueError('Cannot set both "pem" and "b64" fields; exactly one must be specified')
        elif not self.pem and not self.b64:
            raise ValueError('Must set exactly one of "pem" or "b64" fields')
        else:
            return self

    def parse_document(self) -> document.Document:
        pass


class CertificateModel(DocumentModel):
    _parsed_document = None

    @model_validator(mode='after')
    def validate(self) -> 'CertificateModel':
        super()._validate()

        if self.pem is not None:
            try:
                self._parsed_document = loader.load_pem_certificate(self.pem, 'request', 'request')
            except PyAsn1Error as e:
                raise ValueError('Invalid PEM text specified') from e
        else:
            b64 = base64.b64decode(self.b64)

            try:
                self._parsed_document = loader.load_der_certificate(b64, 'request', 'request')
            except PyAsn1Error as e:
                raise ValueError('Invalid Base-64 encoding specified') from e

        return self

    @property
    def parsed_document(self):
        return self._parsed_document



# Auto-generated by asn1ate v.0.6.0 from ts_119_495.asn1
# (last modified on 2023-03-15 09:20:01.431627)

from pyasn1.type import univ, char, namedtype, constraint


def _OID(*components):
    output = []
    for x in tuple(components):
        if isinstance(x, univ.ObjectIdentifier):
            output.extend(list(x))
        else:
            output.append(int(x))

    return univ.ObjectIdentifier(output)


class NCAId(char.UTF8String):
    pass


NCAId.subtypeSpec = constraint.ValueSizeConstraint(1, 256)


class NCAName(char.UTF8String):
    pass


NCAName.subtypeSpec = constraint.ValueSizeConstraint(1, 256)


class RoleOfPspOid(univ.ObjectIdentifier):
    pass


class RoleOfPspName(char.UTF8String):
    pass


RoleOfPspName.subtypeSpec = constraint.ValueSizeConstraint(1, 256)


class RoleOfPSP(univ.Sequence):
    pass


RoleOfPSP.componentType = namedtype.NamedTypes(
    namedtype.NamedType('roleOfPspOid', RoleOfPspOid()),
    namedtype.NamedType('roleOfPspName', RoleOfPspName())
)


class RolesOfPSP(univ.SequenceOf):
    pass


RolesOfPSP.componentType = RoleOfPSP()


class PSD2QcType(univ.Sequence):
    pass


PSD2QcType.componentType = namedtype.NamedTypes(
    namedtype.NamedType('rolesOfPSP', RolesOfPSP()),
    namedtype.NamedType('nCAName', NCAName()),
    namedtype.NamedType('nCAId', NCAId())
)


etsi_psd2_policy = _OID(0, 4, 0, 19495, 3)


etsi_psd2_roles = _OID(0, 4, 0, 19495, 1)


id_etsi_psd2_qcStatement = _OID(0, 4, 0, 19495, 2)


id_psd2_role_psp_ai = _OID(0, 4, 0, 19495, 1, 3)


id_psd2_role_psp_as = _OID(0, 4, 0, 19495, 1, 1)


id_psd2_role_psp_ic = _OID(0, 4, 0, 19495, 1, 4)


id_psd2_role_psp_pi = _OID(0, 4, 0, 19495, 1, 2)


id_psd2_role_psp_unspecified = _OID(0, 4, 0, 19495, 1, 0)


qcp_web_psd2 = _OID(0, 4, 0, 19495, 3, 1)





import binascii

from pyasn1_alt_modules import rfc3279

from pkilint import validation
from pkilint.pkix import algorithm
from pkilint.pkix.certificate import certificate_key

ALLOWED_SIGNATURE_ALGORITHM_ENCODINGS = set(
    map(
        binascii.a2b_hex, [
            # RSASSA‐PKCS1‐v1_5 with SHA‐256
            '300d06092a864886f70d01010b0500',
            # RSASSA‐PKCS1‐v1_5 with SHA‐384
            '300d06092a864886f70d01010c0500',
            # RSASSA‐PKCS1‐v1_5 with SHA‐512
            '300d06092a864886f70d01010d0500',
            # RSASSA‐PSS with SHA‐256, MGF‐1 with SHA‐256, and a salt length of 32 bytes
            '304106092a864886f70d01010a3034a00f300d060960864801650'
            '30402010500a11c301a06092a864886f70d010108300d06096086480165030402010500a203020120',
            # RSASSA‐PSS with SHA‐384, MGF‐1 with SHA‐384, and a salt length of 48 bytes
            '304106092a864886f70d01010a3034a00f300d060960864801650'
            '30402020500a11c301a06092a864886f70d010108300d06096086480165030402020500a203020130',
            # RSASSA‐PSS with SHA‐512, MGF‐1 with SHA‐512, and a salt length of 64 bytes
            '304106092a864886f70d01010a3034a00f300d060960864801650'
            '30402030500a11c301a06092a864886f70d010108300d06096086480165030402030500a203020140',
            # ECDSA with SHA‐256
            '300a06082a8648ce3d040302',
            # ECDSA with SHA‐384
            '300a06082a8648ce3d040303',
            # Ed25519
            '300506032b6570',
            # Ed448
            '300506032b6571'
        ]
    )
)

ALLOWED_SPKI_ENCODINGS = set(
    map(
        binascii.a2b_hex,
        [
            # RSA
            '300d06092a864886f70d0101010500',
            # ECDSA P-256
            '301306072a8648ce3d020106082a8648ce3d030107',
            # ECDSA P-384
            '301006072a8648ce3d020106052b81040022',
            # Ed25519
            '300506032b6570',
            # Ed448
            '300506032b6571'
        ]
    )
)


class SmimeAllowedSignatureAlgorithmEncodingValidator(
    algorithm.AllowedSignatureAlgorithmEncodingValidator
):
    VALIDATION_PROHIBITED_SIGNATURE_ALGORITHM = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'cabf.smime.prohibited_signature_algorithm_encoding'
    )

    def __init__(self, **kwargs):
        super().__init__(
            validation=self.VALIDATION_PROHIBITED_SIGNATURE_ALGORITHM,
            allowed_encodings=ALLOWED_SIGNATURE_ALGORITHM_ENCODINGS,
            **kwargs
        )


class SmimeAllowedPublicKeyAlgorithmEncodingValidator(
    certificate_key.AllowedPublicKeyAlgorithmEncodingValidator
):
    VALIDATION_PROHIBITED_KEY_ALGORITHM = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'cabf.smime.prohibited_spki_algorithm_encoding'
    )

    def __init__(self, **kwargs):
        super().__init__(
            validation=self.VALIDATION_PROHIBITED_KEY_ALGORITHM,
            allowed_encodings=ALLOWED_SPKI_ENCODINGS,
            **kwargs
        )


class GmailAllowedModulusLengthValidator(validation.Validator):
    VALIDATION_PROHIBITED_RSA_MODULUS_LENGTH = validation.ValidationFinding(
        validation.ValidationFindingSeverity.NOTICE,
        'googl.prohibited_rsa_modulus_length'
    )

    _ALLOWED_LENGTHS = {2048, 3072, 4096}

    def __init__(self):
        super().__init__(
            validations=[self.VALIDATION_PROHIBITED_RSA_MODULUS_LENGTH],
            pdu_class=rfc3279.RSAPublicKey
        )

    def validate(self, node):
        modulus_length = int(node.children['modulus'].pdu).bit_length()

        if modulus_length not in self._ALLOWED_LENGTHS:
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_PROHIBITED_RSA_MODULUS_LENGTH,
                f'Prohibited RSA modulus length: {modulus_length}'
            )



from pyasn1_alt_modules import rfc6960

from pkilint import document, validation
from pkilint.itu.bitstring import NamedBitStringMinimalEncodingValidator
from pkilint.itu.string import PrintableStringConstraintValidator
from pkilint.pkix import time
from pkilint.pkix.ocsp import ocsp_response, ocsp_basic_response, ocsp_validity


class RFC6960OCSPResponse(document.Document):
    def __init__(self, substrate_source, substrate,
                 name=None, parent=None
                 ):
        super().__init__(
            rfc6960.OCSPResponse(), substrate_source, substrate, name, parent
        )


def create_response_decoder():
    decoder = ocsp_response.OCSPResponseDecoder(
        type_mappings=rfc6960.ocspResponseMap
    )

    return ocsp_response.OCSPResponseDecodingValidator(decode_func=decoder)


def create_pkix_ocsp_response_validator_container(
        decoding_validators, validators
):
    decoding_validator_containers = [
        validation.ValidatorContainer(
            validators=decoding_validators,
            path='oCSPResponse'
        )
    ]

    validators += [
        ocsp_response.OCSPResponseStatusValidator(),
        ocsp_response.OCSPResponseIsBasicValidator(),
        ocsp_basic_response.OCSPBasicResponseCertsNotPresentValidator(),
        ocsp_basic_response.ResponderKeyHashIsSHA1HashValidator(),
        ocsp_validity.OCSPSaneValidityPeriodValidator(),
        PrintableStringConstraintValidator(),
        time.UtcTimeCorrectSyntaxValidator(),
        time.GeneralizedTimeCorrectSyntaxValidator(),
        NamedBitStringMinimalEncodingValidator(),
    ]

    return validation.ValidatorContainer(
        validators=decoding_validator_containers + validators
    )



from pyasn1.type import namedtype, char, constraint
from pyasn1.type.univ import Choice, ObjectIdentifier

id_at_businessCategory = ObjectIdentifier('2.5.4.15')

ub_business_category = 128


class X520BusinessCategory(Choice):
    pass


X520BusinessCategory.componentType = namedtype.NamedTypes(
    namedtype.NamedType('teletexString',
                        char.TeletexString().subtype(
                            subtypeSpec=constraint.ValueSizeConstraint(1, ub_business_category))),
    namedtype.NamedType('printableString',
                        char.PrintableString().subtype(
                            subtypeSpec=constraint.ValueSizeConstraint(1, ub_business_category))),
    namedtype.NamedType('universalString',
                        char.UniversalString().subtype(
                            subtypeSpec=constraint.ValueSizeConstraint(1, ub_business_category))),
    namedtype.NamedType('utf8String',
                        char.UTF8String().subtype(subtypeSpec=constraint.ValueSizeConstraint(1, ub_business_category))),
    namedtype.NamedType('bmpString',
                        char.BMPString().subtype(subtypeSpec=constraint.ValueSizeConstraint(1, ub_business_category)))
)

id_at_postalCode = ObjectIdentifier('2.5.4.17')

ub_postal_code = 40


class X520PostalCode(Choice):
    pass


X520PostalCode.componentType = namedtype.NamedTypes(
    namedtype.NamedType('teletexString',
                        char.TeletexString().subtype(subtypeSpec=constraint.ValueSizeConstraint(1, ub_postal_code))),
    namedtype.NamedType('printableString',
                        char.PrintableString().subtype(subtypeSpec=constraint.ValueSizeConstraint(1, ub_postal_code))),
    namedtype.NamedType('universalString',
                        char.UniversalString().subtype(subtypeSpec=constraint.ValueSizeConstraint(1, ub_postal_code))),
    namedtype.NamedType('utf8String',
                        char.UTF8String().subtype(subtypeSpec=constraint.ValueSizeConstraint(1, ub_postal_code))),
    namedtype.NamedType('bmpString',
                        char.BMPString().subtype(subtypeSpec=constraint.ValueSizeConstraint(1, ub_postal_code)))
)

id_at_streetAddress = ObjectIdentifier('2.5.4.9')

ub_street_address = 128


class X520StreetAddress(Choice):
    pass


X520StreetAddress.componentType = namedtype.NamedTypes(
    namedtype.NamedType('teletexString',
                        char.TeletexString().subtype(subtypeSpec=constraint.ValueSizeConstraint(1, ub_street_address))),
    namedtype.NamedType('printableString',
                        char.PrintableString().subtype(
                            subtypeSpec=constraint.ValueSizeConstraint(1, ub_street_address))),
    namedtype.NamedType('universalString',
                        char.UniversalString().subtype(
                            subtypeSpec=constraint.ValueSizeConstraint(1, ub_street_address))),
    namedtype.NamedType('utf8String',
                        char.UTF8String().subtype(subtypeSpec=constraint.ValueSizeConstraint(1, ub_street_address))),
    namedtype.NamedType('bmpString',
                        char.BMPString().subtype(subtypeSpec=constraint.ValueSizeConstraint(1, ub_street_address)))
)

id_at_organizationIdentifier = ObjectIdentifier('2.5.4.97')


class X520OrganizationIdentifier(Choice):
    pass


X520OrganizationIdentifier.componentType = namedtype.NamedTypes(
    namedtype.NamedType('teletexString', char.TeletexString()),
    namedtype.NamedType('printableString', char.PrintableString()),
    namedtype.NamedType('universalString', char.UniversalString()),
    namedtype.NamedType('utf8String', char.UTF8String()),
    namedtype.NamedType('bmpString', char.BMPString())
)

ATTRIBUTE_TYPE_MAPPINGS = {
    id_at_businessCategory: X520BusinessCategory(),
    id_at_postalCode: X520PostalCode(),
    id_at_streetAddress: X520StreetAddress(),
    id_at_organizationIdentifier: X520OrganizationIdentifier(),
}



from pyasn1.type import univ, char, constraint

MAX = float('inf')

id_on_UserPrincipalName = univ.ObjectIdentifier('1.3.6.1.4.1.311.20.2.3')


class UserPrincipalName(char.UTF8String):
    subtypeSpec = constraint.ValueSizeConstraint(1, MAX)



from pyasn1.codec.der.encoder import encode
from pyasn1.type.univ import BitString

from pkilint import validation


class NamedBitStringMinimalEncodingValidator(validation.Validator):
    VALIDATION_BIT_STRING_NOT_MINIMALLY_ENCODED = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'itu.bitstring_not_der_encoded'
    )

    def __init__(self):
        super().__init__(
            validations=[self.VALIDATION_BIT_STRING_NOT_MINIMALLY_ENCODED],
            pdu_class=BitString,
            predicate=lambda n: any(n.pdu.namedValues)
        )

    def validate(self, node):
        # extract values then re-encode

        asserted_values = ','.join((k for k in node.pdu.namedValues.keys() if has_named_bit(node, k)))

        encoded = encode(node.pdu)

        new_encoded = encode(type(node.pdu)(asserted_values), asn1Spec=node.pdu)

        if encoded != new_encoded:
            encoded_hex = encoded.hex()
            new_encoded_hex = new_encoded.hex()

            raise validation.ValidationFindingEncountered(
                self.VALIDATION_BIT_STRING_NOT_MINIMALLY_ENCODED,
                f'Expected: "{new_encoded_hex}", actual: "{encoded_hex}"'
            )


def has_named_bit(node, bit_name):
    bit = node.pdu.namedValues[bit_name]
    return len(node.pdu) > bit and node.pdu[bit] != 0



from cryptography.hazmat.primitives import hashes
from pyasn1_alt_modules import rfc6960

from pkilint import validation


class OCSPBasicResponseCertsNotPresentValidator(validation.Validator):
    VALIDATION_CERTS_IS_EMPTY = validation.ValidationFinding(
        validation.ValidationFindingSeverity.WARNING,
        'pkix.ocsp_certs_sequence_is_empty'
    )

    def __init__(self):
        super().__init__(
            validations=[self.VALIDATION_CERTS_IS_EMPTY],
            pdu_class=rfc6960.BasicOCSPResponse
        )

    def validate(self, node):
        certs_node = node.children.get('certs')

        if certs_node is not None and len(certs_node.children) == 0:
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_CERTS_IS_EMPTY
            )


class ResponderKeyHashIsSHA1HashValidator(validation.Validator):
    VALIDATION_KEY_HASH_IS_NOT_SHA1 = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        'pkix.ocsp_responderid_keyhash_is_not_sha1'
    )

    def __init__(self):
        super().__init__(validations=[self.VALIDATION_KEY_HASH_IS_NOT_SHA1],
                         pdu_class=rfc6960.ResponderID,
                         predicate=lambda n: 'byKey' in n.children
                         )

    def validate(self, node):
        _, by_key_node = node.child

        hash_octets = by_key_node.pdu.asOctets()
        hash_len = len(hash_octets)

        if hash_len != hashes.SHA1.digest_size:
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_KEY_HASH_IS_NOT_SHA1,
                f'Key hash length of {hash_len} octets is not SHA-1'
            )



import enum

from pyasn1_alt_modules import rfc5280

from pkilint import validation, pkix, document
from pkilint.document import Document
from pkilint.itu.bitstring import NamedBitStringMinimalEncodingValidator
from pkilint.itu.string import PrintableStringConstraintValidator
from pkilint.pkix import name, extension, time, general_name
from pkilint.pkix.crl import crl_validator, crl_extension, crl_validity


@enum.unique
class CertificateRevocationListType(enum.IntEnum):
    CRL = 0
    ARL = 1


class RFC5280CertificateList(Document):
    def __init__(self, substrate_source, substrate,
                 name=None, parent=None
                 ):
        super().__init__(
            rfc5280.CertificateList(), substrate_source, substrate, name, parent
        )

    @property
    def this_update(self):
        return time.parse_time_node(
            self.root.navigate('tbsCertificateList.thisUpdate')
        )

    @property
    def next_update(self):
        try:
            return time.parse_time_node(
                self.root.navigate('tbsCertificateList.nextUpdate')
            )
        except document.PDUNavigationFailedError:
            return None

    def get_extension_by_oid(self, oid):
        tbs_crl = self.root.children['tbsCertList']

        # ensure there's extensions
        if 'crlExtensions' not in tbs_crl.children:
            return None

        extensions_node = tbs_crl.children['crlExtensions']

        for ext_idx, extension_node in extensions_node.children.items():
            ext_oid = extension_node.children['extnID'].pdu

            if ext_oid == oid:
                return extension_node, int(ext_idx)

        return None


def create_issuer_validator_container(additional_validators=None):
    if additional_validators is None:
        additional_validators = []
    additional_validators.append(name.EmptyNameValidator())

    return pkix.create_name_validator_container(
         additional_validators, path='certificateList.tbsCertList.issuer'
    )


def create_validity_validator_container(additional_validators=None):
    if additional_validators is None:
        additional_validators = []
    return validation.ValidatorContainer(
        validators=[
                       crl_validity.CrlSaneValidityPeriodValidator(),
                       time.TimeCorrectEncodingValidator(),
                   ] + additional_validators,
        path='certificateList.tbsCertList'
    )


def create_extensions_validator_container(additional_validators=None):
    if additional_validators is None:
        additional_validators = []
    return validation.ValidatorContainer(
        validators=[
                       # extension.PermittedExtensionValidator(
                       #    known_oids=known_oids
                       # ),
                       extension.UniqueExtensionValidator(),
                       crl_extension.CrlNumberCriticalityValidator(),
                       extension.AuthorityKeyIdentifierValidator(),
                   ] + additional_validators,
        path='certificateList.tbsCertList.crlExtensions'
    )


def create_pkix_crl_validator_container(
        decoding_validators, validators):
    decoding_validator_container = [
        validation.ValidatorContainer(
            validators=decoding_validators, path='certificateList'
        )
    ]

    validators += [
        PrintableStringConstraintValidator(),
        crl_validator.VersionPresenceValidator(),
        crl_validator.CorrectVersionValidator(),
        crl_extension.CrlNumberPresenceValidator(),
        crl_extension.AuthorityKeyIdentifierPresenceValidator(),
        crl_validator.SignatureAlgorithmMatchValidator(),
        crl_extension.CrlReasonCodeCriticalityValidator(),
        time.UtcTimeCorrectSyntaxValidator(),
        time.GeneralizedTimeCorrectSyntaxValidator(),
        pkix.CertificateSerialNumberValidator(),
        crl_extension.CrlNumberValueValidator(),
        NamedBitStringMinimalEncodingValidator(),
        general_name.GeneralNameIpAddressSyntaxValidator(),
        general_name.GeneralNameMailboxAddressSyntaxValidator(),
        general_name.GeneralNameIpAddressSyntaxValidator(),
        general_name.GeneralNameUriSyntaxValidator(),
    ]

    return validation.ValidatorContainer(
        validators=decoding_validator_container + validators
    )



from pyasn1.type import univ, namedtype, namedval
from pyasn1_alt_modules import rfc5280


_ADOBE_X509_OID_ARC = univ.ObjectIdentifier('1.2.840.113583.1.1.9')


id_adobe_timestamp = univ.ObjectIdentifier(_ADOBE_X509_OID_ARC.asTuple() + (1,))


class AdobeExtensionVersion(univ.Integer):
    pass

AdobeExtensionVersion.componentType = namedval.NamedValues(
    ('v1', 1),
)


class AdobeTimestamp(univ.Sequence):
    pass


AdobeTimestamp.componentType = namedtype.NamedTypes(
    namedtype.NamedType('version', AdobeExtensionVersion()),
    namedtype.NamedType('location', rfc5280.GeneralName()),
    namedtype.DefaultedNamedType('requiresAuth', univ.Boolean().subtype(value=False))
)


id_adobe_archiverevinfo = univ.ObjectIdentifier(_ADOBE_X509_OID_ARC.asTuple() + (2,))


class AdobeArchiveRevInfo(univ.Sequence):
    pass


AdobeArchiveRevInfo.componentType = namedtype.NamedTypes(
    namedtype.NamedType('version', AdobeExtensionVersion())
)


EXTENSION_MAPPINGS = {
    id_adobe_timestamp: AdobeTimestamp(),
    id_adobe_archiverevinfo: AdobeArchiveRevInfo(),
}



from typing import List, Sequence, Tuple, NamedTuple

from pkilint.validation import ValidationResult, ValidationFinding, ValidationFindingDescription


class FindingDescriptionFilter:
    def __init__(self):
        pass

    def filter(self, result: ValidationResult, finding_description: ValidationFindingDescription) -> bool:
        pass


class ResultWithExcludedFindingDescriptions(NamedTuple):
    result: ValidationResult
    filter_and_finding_descriptions: List[Tuple[FindingDescriptionFilter, ValidationFindingDescription]]


class ValidationFindingFilter(FindingDescriptionFilter):
    def __init__(self, validation_finding: ValidationFinding):
        self._validation_finding = validation_finding

        super().__init__()

    def filter(self, result: ValidationResult, finding_description: ValidationFindingDescription) -> bool:
        return self._validation_finding != finding_description.finding


def filter_results(filters: List[FindingDescriptionFilter], results: Sequence[ValidationResult]) -> Tuple[
        List[ValidationResult], List[ResultWithExcludedFindingDescriptions]]:
    filtered_results = []
    results_with_exclusions = []

    for result in results:
        included_finding_descriptions = []
        filtered_finding_descriptions = []

        for finding_description in result.finding_descriptions:
            excluding_filter = next((f for f in filters if not f.filter(result, finding_description)), None)

            if excluding_filter is None:
                included_finding_descriptions.append(finding_description)
            else:
                filtered_finding_descriptions.append((excluding_filter, finding_description,))

        if any(filtered_finding_descriptions):
            results_with_exclusions.append(ResultWithExcludedFindingDescriptions(
                ValidationResult(result.validator, result.node, [fafd[1] for fafd in filtered_finding_descriptions]),
                filtered_finding_descriptions
            ))

            filtered_results.append(ValidationResult(result.validator, result.node, included_finding_descriptions))
        else:
            filtered_results.append(result)

    return filtered_results, results_with_exclusions



import binascii
import logging
import re
from typing import Callable, Mapping, Tuple, Type, Union, Optional, Dict, List

from pyasn1.codec.der.decoder import decode
from pyasn1.codec.der.encoder import encode
from pyasn1.error import PyAsn1Error
from pyasn1.type.base import Asn1Type
from pyasn1.type.univ import (ObjectIdentifier, SequenceOfAndSetOfBase, SequenceAndSetBase,
                              Choice, BitString
                              )

logger = logging.getLogger(__name__)

PATH_REGEX = re.compile(r'^((?P<doc_name>[^:]*):)?(?P<node_path>([^.]+\.)*[^.]+)?$')


class PDUNavigationFailedError(Exception):
    """Represents the failure to find the requested node in a document."""

    def __init__(self, requested_path: str, traversed_path: str,
                 missing_node_name: str
                 ):
        """Creates an instance of an exception that represents a PDU node lookup failure.

        Args: requested_path: The requested path relative to the node which
        :py:method:`pkilint.document.PDUNode.navigate was called. traversed_path: The relative path that was able to
        be traversed, as those nodes exist. missing_node_name: The name of the node which could not be found.
        """
        self.requested_path = requested_path
        self.traversed_path = traversed_path
        self.missing_node_name = missing_node_name

    def __str__(self) -> str:
        return (f'Node with name "{self.missing_node_name}" does not exist at '
                f'"{self.traversed_path}" (requested path: "{self.requested_path}")'
                )


class Document(object):
    """Represents an ASN.1-encoded document."""

    def __init__(
            self, pdu_schema_instance: Asn1Type, substrate_source: str, substrate: bytes,
            name: Optional[str] = None,
            parent: Optional[Mapping[str, 'Document']] = None
    ):
        """Creates a new Document instance. It is not intended that this class be directly
        instantiated by user code; use a sub-class of this class instead.

        Args:
            pdu_schema_instance: A pyasn1 ASN.1 instance that represents the top-level ASN.1 schema for this document.
            substrate_source: The source of the document. This can be a URI, file name, or any other identifier.
            substrate: The raw DER-encoded document.
            name: An optional name given to the document. May be useful when a :py:meth:`pkilint.validation.Validator` requires multiple documents.
            parent: An optional collection of documents that are related.
        """
        self.pdu_schema_instance = pdu_schema_instance
        self.substrate = substrate
        self.substrate_source = substrate_source

        self.name = name
        self.parent = parent
        self.root = None

    def decode(self):
        """
            Decodes the DER-encoded substrate with the specified ASN.1 schema object.

            If the document does not conform to the schema, then this will fail.
        """
        if self.root is None:
            self.root = decode_substrate(self, self.substrate,
                                         self.pdu_schema_instance
                                         )

        return self.root

    def __repr__(self):
        return f'{self.root.name} document "{self.substrate_source}"'


class PDUNode(object):
    """Represents a node of a document."""

    def __init__(self, document: Document, name: str, pdu: Asn1Type,
                 parent: Optional['PDUNode']
                 ):
        """Creates a new instance representing a node within a document.

        Args:
            document: The document which contains this node.
            name: The name of the node. Generally will match the name of a component
            within an ASN.1 SEQUENCE.
            pdu: The underlying ASN.1 value.
            parent: The node which contains this node. In the case where the current node is the top-level node of a document, this will not be populated.
        """
        self.document = document
        self.name = name
        self.pdu = pdu
        self.parent = parent
        if self.parent is None:
            self.path = self.name
        else:
            self.path = f'{self.parent.path}.{self.name}'

        self.children = self._generate_child_nodes()

    @property
    def parents(self) -> List['PDUNode']:
        """All parent nodes up to the root of the document."""
        nodes = []
        node = self.parent
        while node is not None:
            nodes.append(node)
            node = node.parent

        return nodes

    @property
    def child(self) -> Tuple[str, 'PDUNode']:
        """The node name and node of the single child node.

        This property will fail if no child nodes or more than one child node is present.
        """
        if len(self.children) > 1:
            raise ValueError(f'"{self}" has multiple child nodes')
        elif len(self.children) == 0:
            raise ValueError(f'"{self}" has no child nodes')
        else:
            return next(iter(self.children.items()))

    def navigate(self, path: str) -> Union['PDUNode', Document]:
        """Navigates to a node or document (depending on the path specified).

        Elements within a path are separated by periods ("."). Paths may be absolute or relative. Relative paths can
        be relative to the current document or the current node, depending on the syntax used.

        Absolute paths begin with the document name followed by a colon (":"). If no sub-elements are specified,
        then the complete document matching the specified name is returned. For example: `subject:` returns the
        document named "subject". Absolute paths may also contain a node path following the document name and colon.

        Relative paths consist of a leading colon (which anchors the path to the root of the current document) and/or
        a node path.

        The node path element "^" navigates to the parent node.

        Args:
            path: The requested path.
        """
        requested_path = path

        m = PATH_REGEX.match(path)

        if m is None:
            raise ValueError(f'Invalid path syntax: "{path}"')

        doc_name = m.group('doc_name')
        if m.group('node_path') is None:
            node_path_parts = []
        else:
            node_path_parts = m.group('node_path').split('.')

        if doc_name is None:
            node = self
        else:
            if doc_name == '':
                doc = self.document

                if len(node_path_parts) == 0:
                    return doc
            else:
                if self.document.parent is None or doc_name not in self.document.parent:
                    raise PDUNavigationFailedError(requested_path, '', doc_name)

                doc = self.document.parent[doc_name]

            root_name = node_path_parts[0]
            if doc.root.name != root_name:
                raise PDUNavigationFailedError(requested_path, doc_name, root_name)

            node = doc.root
            node_path_parts = node_path_parts[1:]

        for part in node_path_parts:
            if part == '^':
                node = node.parent
            else:
                try:
                    node = node.children[part]
                except KeyError:
                    raise PDUNavigationFailedError(requested_path, node.path, part)

        return node

    def _generate_child_nodes(self):
        if isinstance(self.pdu, Choice):
            name = self.pdu.getName()
            return {name: PDUNode(
                self.document, name, self.pdu.getComponent(), self
            )
            }
        elif isinstance(self.pdu, SequenceOfAndSetOfBase):
            return {
                str(i): PDUNode(
                    self.document, str(i), component, self
                )
                for i, component in enumerate(self.pdu)
            }
        elif isinstance(self.pdu, SequenceAndSetBase):
            return {name: PDUNode(self.document, name, value, self)
                    for name, value in self.pdu.items()
                    if value.isValue
                    }
        else:
            return {}

    def __repr__(self):
        if self.document is not None and self.document.name is not None:
            path = f'{self.document.name}:{self.path}'
        else:
            path = self.path
        return f'{self.pdu.__class__.__name__} @ {path}'


class NodeVisitor(object):
    def __init__(self, *,
                 path: str = None,
                 path_re: re.Pattern = None,
                 pdu_class: Type[Asn1Type] = None,
                 pdu_supertype: Asn1Type = None,
                 predicate: Callable[[PDUNode], bool] = None
                 ):
        self._path = path
        self._path_re = path_re
        self._pdu_class = pdu_class
        self._pdu_supertype = pdu_supertype
        self._predicate = predicate

    def match(self, node: PDUNode) -> bool:
        if self._path is not None and self._path != node.path:
            return False
        if self._path_re is not None and self._path_re.match(node.path) is None:
            return False
        if self._pdu_class is not None and not isinstance(node.pdu, self._pdu_class):
            return False
        if self._pdu_supertype is not None and not self._pdu_supertype.isSuperTypeOf(node.pdu):
            return False
        if self._predicate is not None and not self._predicate(node):
            return False

        return True


class ValueDecodingFailedError(Exception):
    def __init__(self, value_node: PDUNode, type_oid: ObjectIdentifier,
                 pdu_type: Optional[Asn1Type], message: str
                 ):
        self.value_node = value_node
        self.type_oid = type_oid
        self.pdu_type = pdu_type
        self.message = message


class ValueDecoder(object):
    _BITSTRING_SCHEMA_OBJ = BitString()

    VALUE_NODE_ABSENT = object()

    def __init__(self, *, type_path: str, value_path: str,
                 type_mappings: Dict[ObjectIdentifier, Optional[Asn1Type]], default: Optional[Asn1Type] = None):
        self.type_path = type_path
        self.value_path = value_path
        self.type_mappings = type_mappings.copy()
        self.default = default

    def filter_value(self, node, type_node, value_node, pdu_type):
        if self._BITSTRING_SCHEMA_OBJ.isSuperTypeOf(value_node.pdu):
            return value_node.pdu.asOctets()
        else:
            return value_node.pdu

    def __call__(self, node):
        type_node = node.navigate(self.type_path)

        try:
            value_node = node.navigate(self.value_path)
        except PDUNavigationFailedError:
            value_node = None

        pdu_type = self.type_mappings.get(type_node.pdu, self.default)

        if pdu_type is not None:
            # value node must be absent, but it exists
            if pdu_type is self.VALUE_NODE_ABSENT and value_node is not None:
                raise ValueDecodingFailedError(
                    value_node, type_node.pdu, pdu_type,
                    'Value node is present, but the ASN.1 schema specifies that it must be absent'
                )
            # value node must be present, but it doesn't exist
            if pdu_type is not self.VALUE_NODE_ABSENT and value_node is None:
                raise ValueDecodingFailedError(
                    node, type_node.pdu, pdu_type,
                    'Value node is absent, but the ASN.1 schema specifies that it must be present'
                )

        if pdu_type is self.VALUE_NODE_ABSENT or pdu_type is None:
            return

        value_octets = self.filter_value(node, type_node, value_node, pdu_type)

        try:
            decode_substrate(value_node.document, value_octets,
                             pdu_type, value_node
                             )
        except (PyAsn1Error, ValueError) as e:
            logger.exception(
                'Error occurred when decoding %s value in %s',
                pdu_type,
                value_node.path
            )

            raise ValueDecodingFailedError(
                value_node, type_node.pdu, pdu_type, str(e)
            )


def get_node_name_for_pdu(pdu: Asn1Type) -> str:
    name = pdu.__class__.__name__
    # convert PDU class name to camelCase
    return name[0].lower() + name[1:]


def get_document_by_name(node: PDUNode, document_name: str) -> Document:
    """Retrieves the document with the specified name"""
    return node.document.parent[document_name]


def decode_substrate(source_document: Document, substrate: bytes,
                     pdu_instance: Asn1Type, parent_node: Optional[PDUNode] = None) -> PDUNode:
    if parent_node is not None and len(parent_node.children) > 0:
        logger.debug("%s has child node; not creating new PDU node",
                     parent_node.path
                     )
        return next(iter(parent_node.children.values()))

    decoded, rest = decode(substrate, asn1Spec=pdu_instance)

    decoded_pdu_name = get_node_name_for_pdu(decoded)

    if len(rest) > 0:
        raise ValueError(
            "Unexpected {} octets following {} DER in {}: {}".format(
                len(rest), decoded_pdu_name, source_document.substrate_source,
                binascii.hexlify(rest).decode('us-ascii')
            )
        )

    encoded = encode(decoded)
    if encoded != substrate:
        type_name = decoded.__class__.__name__
        raise ValueError(
            f'Substrate of type "{type_name}" is not DER-encoded'
        )

    node = PDUNode(source_document, decoded_pdu_name, decoded, parent_node)

    if parent_node is not None:
        parent_node.children[decoded_pdu_name] = node
        logger.debug("Appended %s node to %s", node.name,
                     parent_node.path
                     )

    return node


def get_re_for_path_glob(path_glob: str) -> re.Pattern:
    return re.compile(
        path_glob.replace('.', r'\.').replace('?', r'\w').replace('*', r'\w*')
    )



import enum
import itertools
import logging
from typing import Any, Callable, Mapping, NamedTuple, List, Tuple, Optional

from pyasn1.codec.der.encoder import encode
from pyasn1.type.constraint import PermittedAlphabetConstraint, ValueRangeConstraint
from pyasn1.type.error import ValueConstraintError
from pyasn1.type.univ import ObjectIdentifier

from pkilint.document import (PDUNode, NodeVisitor, ValueDecodingFailedError,
                              PDUNavigationFailedError
                              )

logger = logging.getLogger(__name__)


@enum.unique
class ValidationFindingSeverity(enum.IntEnum):
    """Represents the severity of a finding raised by a validator"""

    FATAL = 0
    '''A violation so severe that the validator cannot proceed'''

    ERROR = 1
    '''A clear violation of a standard or policy (e.g., violating a "MUST")'''

    WARNING = 2
    '''A clear violation of best practice (e.g., violating a "SHOULD")'''

    NOTICE = 3
    '''Not a clear violation of best practice or standard, but something to be noted'''

    INFO = 4
    '''Informative findings that are not indicative of any problem'''

    DEBUG = 5
    '''Information solely useful for debugging'''

    def __str__(self):
        return self.name


class ValidationFinding(NamedTuple):
    """Represents a finding that may be raised by a validator"""

    severity: ValidationFindingSeverity
    '''The severity of the validation finding'''

    code: str
    '''An identifier that specifies the part of standard or policy that is being validated'''

    def __repr__(self):
        return f'{self.code} ({self.severity.name})'


class ValidationFindingDescription(NamedTuple):
    """Represents a finding that is raised by a validator. Optionally may include a message"""

    finding: ValidationFinding
    '''The finding raised by the validator'''

    message: Optional[str]
    '''A optional message that provides additional human-readable context for the finding'''

    def __repr__(self):
        message = f': {self.message}' if self.message else ''

        return f'{self.finding}{message}'


class Validator(NodeVisitor):
    """Validates a document (or part thereof) for compliance with a policy or standard"""

    VALIDATION_FINDING_UNHANDLED_EXCEPTION = ValidationFinding(
        ValidationFindingSeverity.FATAL,
        'base.unhandled_exception'
    )
    '''A finding that indicates an exception occurred during validation'''

    def __init__(self, *, validations=None, **kwargs):
        super().__init__(**kwargs)

        if validations is None:
            validations = []
        if not isinstance(validations, list):
            validations = [validations]

        self._validations = validations

    def validate_wrapper(self, node: PDUNode) -> 'ValidationResult':
        try:
            # pylint: disable=assignment-from-no-return
            results = self.validate(node)
            if results is None:
                return ValidationResult(self, node, [])
            else:
                return results

        except ValidationFindingEncountered as e:
            finding = ValidationFindingDescription(e.finding, e.message)
        except Exception as e:
            logger.exception('Unhandled exception occurred when executing '
                             'validator %s on node %s', self.name, node.path
                             )
            finding = ValidationFindingDescription(
                self.VALIDATION_FINDING_UNHANDLED_EXCEPTION,
                str(e)
            )

        return ValidationResult(self, node, [finding])

    def validate(self, node: PDUNode) -> 'ValidationResult':
        """Validates the specified node"""
        pass

    @property
    def tags(self):
        return ['static']

    @property
    def validations(self):
        return self._validations + [self.VALIDATION_FINDING_UNHANDLED_EXCEPTION]

    @property
    def name(self):
        return self.__class__.__name__

    def __repr__(self):
        return self.name


class ValidationResult(NamedTuple):
    """Represents the result of running a validator on a document node"""
    validator: Validator
    '''The validator that was executed'''
    node: PDUNode
    '''The node of the document that was validated'''
    finding_descriptions: List[ValidationFindingDescription]
    '''The list of findings and their associated messages'''

    def __repr__(self):
        findings_str = ', '.join([str(f) for f in self.finding_descriptions])
        return f'{self.validator} result for "{self.node}": {findings_str}'


class ValidationFindingEncountered(Exception):
    """A convenient way to raise findings from a validator"""

    def __init__(self, finding: ValidationFinding, message: str = None):
        self.finding = finding
        '''The finding that is being raised by the validator'''
        self.message = message
        '''An optional message providing additional human-readable context for the finding'''


class ValidatorContainer(Validator):
    """A collection of validators that recursively executes all included
    validators on the matching document node and its children"""

    def __init__(self, *, validators: List[Validator], **kwargs):
        self.validators = validators
        validations_2d = (v.validations for v in self.validators)
        validations_1d = itertools.chain.from_iterable(validations_2d)

        all_validations = [
            v
            for v in validations_1d
            if v is not self.VALIDATION_FINDING_UNHANDLED_EXCEPTION
        ]

        super().__init__(validations=all_validations, **kwargs)

    def _validate_rec(self, node: PDUNode,
                      results: List[ValidationResult]
                      ):
        for v in self.validators:
            if v.match(node):
                result = v.validate_wrapper(node)

                if isinstance(result, list):
                    results += result
                else:
                    results.append(result)

        for child_node in node.children.values():
            self._validate_rec(child_node, results)

    def validate(self, node: PDUNode) -> List[ValidationResult]:
        results = []

        self._validate_rec(node, results)

        return results


class ScalarFieldValueEqualityValidator(Validator):
    def __init__(self, *, value, **kwargs):
        super().__init__(**kwargs)

        self.value = value

    def validate(self, node: PDUNode):
        if node.pdu != self.value:
            validation = self.validations[0]

            raise ValidationFindingEncountered(
                validation,
                f'Expected="{self.value}", actual="{node.pdu}"'
            )


class ASN1ConstraintValidator(Validator):
    def __init__(self, constraint, **kwargs):
        super().__init__(**kwargs)

        self.constraint = constraint

    def _get_message(self, exc):
        if isinstance(self.constraint, PermittedAlphabetConstraint):
            return ('Invalid character outside permitted alphabet of "'
                    f'{"".join(self.constraint._values)}"'
                    )
        elif isinstance(self.constraint, ValueRangeConstraint):
            return ('Invalid value outside range '
                    f'{self.constraint.start} - {self.constraint.stop}'
                    )
        else:
            return 'ASN.1 constraint violation'

    def validate(self, node):
        try:
            self.constraint(node.pdu)
        except ValueConstraintError as e:
            validation = self.validations[0]

            raise ValidationFindingEncountered(
                validation,
                f'ASN.1 constraint failed: {self._get_message(e)} '
                f'on content "{node.pdu}"'
            )


class DecodingValidator(Validator):
    VALIDATION_ASN1_DECODING_FAILURE = ValidationFinding(
        ValidationFindingSeverity.FATAL,
        'itu.invalid_asn1_syntax'
    )

    def __init__(self, *, decode_func=Callable[[PDUNode], None], **kwargs):
        self.decode_func = decode_func

        super().__init__(validations=[self.VALIDATION_ASN1_DECODING_FAILURE],
                         **kwargs
                         )

    def validate(self, node):
        try:
            self.decode_func(node)
        except ValueDecodingFailedError as e:
            schema_name = e.pdu_type.__class__.__name__
            path = e.value_node.path

            description = (f'ASN.1 decoding failure occurred at "{path}"'
                           f' with schema "{schema_name}", OID {str(e.type_oid)}: '
                           f'{e.message}'
                           )

            raise ValidationFindingEncountered(
                self.VALIDATION_ASN1_DECODING_FAILURE,
                description
            )


class TypeMatchingValidator(Validator):
    def __init__(self, *, type_path: str, type_oid: ObjectIdentifier,
                 value_path: str, **kwargs
                 ):
        self.type_path = type_path
        self.type_oid = type_oid
        self.value_path = value_path

        super().__init__(**kwargs)

    def match(self, node):
        if not super().match(node):
            return False

        type_node = node.navigate(self.type_path)

        return type_node.pdu == self.type_oid

    def validate_with_value(self, node, value_node):
        pass

    def validate(self, node):
        value_node = node.navigate(self.value_path)

        return self.validate_with_value(node, value_node)


class DEREqualityValidator(Validator):
    def __init__(self, *, other_node_retriever=Callable[[PDUNode], PDUNode],
                 validation: ValidationFinding, **kwargs
                 ):
        self._other_node_retriever = other_node_retriever

        super().__init__(
            **kwargs,
            validations=[validation]
        )

    def validate(self, node):
        other_node = self._other_node_retriever(node)

        if encode(node.pdu) != encode(other_node.pdu):
            raise ValidationFindingEncountered(
                self.validations[0],
                f'DER encoding of {node.path} and {other_node.path} are '
                f'not equal'
            )


class NodePresenceValidator(Validator):
    def __init__(self, *, node_retriever: Callable[[PDUNode], PDUNode],
                 absence_finding: ValidationFinding = None,
                 presence_finding: ValidationFinding = None, **kwargs
                 ):
        self._node_retriever = node_retriever
        self._absence_finding = absence_finding
        self._presence_finding = presence_finding

        validations = []
        if absence_finding is not None:
            validations.append(absence_finding)
        if presence_finding is not None:
            validations.append(presence_finding)

        super().__init__(validations=validations, **kwargs)

    def validate(self, node):
        try:
            self._node_retriever(node)

            if self._presence_finding is not None:
                raise ValidationFindingEncountered(self._presence_finding)
        except PDUNavigationFailedError:
            if self._absence_finding is not None:
                raise ValidationFindingEncountered(self._absence_finding)



import functools
import glob
import sys
from os import path

from pkilint.pkix import certificate, name, extension
from tests import integration_certificate

cur_dir = path.dirname(__file__)
test_dir = path.join(cur_dir, 'pkix')
this_module = sys.modules[__name__]

files = glob.glob(path.join(test_dir, '*.crttest'))

for file in files:
    validator = certificate.create_pkix_certificate_validator_container(
        certificate.create_decoding_validators(name.ATTRIBUTE_TYPE_MAPPINGS, extension.EXTENSION_MAPPINGS),
        [
            certificate.create_issuer_validator_container(
                []
            ),
            certificate.create_validity_validator_container(),
            certificate.create_subject_validator_container(
                []
            ),
            certificate.create_extensions_validator_container(
                []
            ),
        ]
    )

    file_no_ext, _ = path.splitext(path.basename(file))

    func_name = f'test_{file_no_ext}'

    setattr(this_module, func_name, functools.partial(integration_certificate.run_test, file, validator))



import csv
import io
import typing
from os import path
from pathlib import Path

from pkilint import loader, validation, finding_filter as result_filter

_FIXTURE_DIR = Path(__file__).parent.resolve()

_CERT_END_ASCII_ARMOR = '-----END CERTIFICATE-----'


class TestFinding(typing.NamedTuple):
    node_path: str
    validator: str
    severity: validation.ValidationFindingSeverity
    code: str
    message: typing.Optional[str]

    @staticmethod
    def from_dict(d):
        return TestFinding(d['node_path'], d['validator'], validation.ValidationFindingSeverity[d['severity']],
                           d['code'], None if not d['message'] else d['message'])

    @staticmethod
    def from_result_and_finding_description(result: validation.ValidationResult,
                                            finding_description: validation.ValidationFindingDescription):
        return TestFinding(
            result.node.path,
            str(result.validator),
            finding_description.finding.severity,
            finding_description.finding.code,
            finding_description.message
        )

    def __repr__(self):
        with io.StringIO() as s:
            c = csv.writer(s)

            c.writerow([self.node_path, self.validator, str(self.severity), self.code, self.message])

            return s.getvalue()

    def __str__(self):
        return repr(self)


def certificate_test_file(file_path):
    with open(file_path, 'r', encoding='utf-8') as f:
        lines = [l.strip() for l in f.readlines() if l.strip() != '']

    assert any((_CERT_END_ASCII_ARMOR in line for line in lines))

    pem_text = ''

    line_num = -1
    for line_num, line in enumerate(lines):
        pem_text += line

        if _CERT_END_ASCII_ARMOR in line:
            break

    with io.StringIO('\n'.join(lines[line_num + 1:])) as s:
        c = csv.DictReader(s)

        expected_findings = set((TestFinding.from_dict(row) for row in c))

        cert = loader.load_pem_certificate(pem_text, path.basename(file_path))
        cert.decode()

    return cert, expected_findings


def run_test(test_file_path, validator, filters=None):
    if filters is None:
        filters = []

    cert, expected_findings = certificate_test_file(test_file_path)

    results = validator.validate(cert.root)
    results, _ = result_filter.filter_results(filters, results)

    actual_findings = set()

    for result in results:
        finding_descriptions = [fd for fd in result.finding_descriptions]

        for finding_description in finding_descriptions:
            actual_findings.add(TestFinding.from_result_and_finding_description(result, finding_description))

    missing_findings = expected_findings - actual_findings
    unexpected_findings = actual_findings - expected_findings

    assert not any(missing_findings) and not any(unexpected_findings), (
        f'Missing findings: {missing_findings}\nUnexpected findings: {unexpected_findings}')
